package basics.datagraph;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Set;
import java.util.TreeSet;

/**
 * This class is part of the YAGO extractors (http://mpii.de/yago). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share Alike 3.0 Unported License,
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) 
 * by the YAGO team (http://mpii.de/yago). 
 * 
 * This class represents a YAGO relation.<P>
 *
 * If a property of relations is added, all methods in the class Relation.java need to be adjusted!
 * RelationExtractor.java does not need to be adjusted.
 * Relations are represented as Java-objects in order to allow other Generators/Extractors to
 * access the information about relations.
 * 
 *  @author Fabian M. Suchanek
 *  
 *  TODO: I have relaxed some ranges (marked) because else we find too few facts
 *  */
public class Relation implements Comparable<Relation> {

  /** Holds the name of this relation*/
  public String name;

  /** Holds the sentence of this relation*/
  public String sentence = null;

  /** Holds the domain of this relation*/
  public YagoClass domain = null;

  /** Holds the range of this relation*/
  public YagoClass range = null;

  /** Points to the super relation (or null)*/
  public Relation superRelation;

  /** Holds the types of this relation */
  public Set<YagoClass> types = null;

  public Relation(String name, String sentence, YagoClass domain, YagoClass range, Relation superRelation, Collection<YagoClass> types) {
    this.name = name;
    this.sentence = sentence;
    this.domain = domain;
    this.range = range;
    this.superRelation = superRelation;
    this.types = new TreeSet<YagoClass>(types);
  }

  public Relation(String name, YagoClass domain, YagoClass range, Collection<YagoClass> types) {
    this(name, domain, range, null, types);
  }

  public Relation(String name, YagoClass domain, YagoClass range, YagoClass type) {
    this(name, domain, range, null, Arrays.asList(type));
  }

  public Relation(String name, YagoClass domain, YagoClass range, Relation superRelation, Collection<YagoClass> types) {
    this(name, null, domain, range, superRelation, types);
  }

  public Relation(String name, YagoClass domain, YagoClass range) {
    this(name, domain, range, null, new ArrayList<YagoClass>(0));
  }

  /** Maps a relation of propertyList to the appropriate value*/
  public Object getProperty(Relation prop) {
    if (prop == Relations.DOMAIN) return domain;
    if (prop == Relations.RANGE) return range;
    if (prop == Relations.SUBPROPERTYOF) return (superRelation == null ? null : superRelation);
    if (prop == Relations.TYPE) {
      if (isTransitive()) return YagoClasses.TRANSITIVERELATION;
      if (isFunction()) return YagoClasses.FUNCTION;
      return (YagoClasses.RELATION);
    }
    return (null);
  }

  /** Says whether this relation is an instance of a class*/
  public boolean isInstanceOf(YagoClass c) {
    return types.contains(c);
  }

  /** Returns the super-Relation*/
  public Relation superRelation() {
    return (superRelation);
  }

  /** Returns the relation name */
  public String toString() {
    return name();
  }

  /** Returns the relation name */
  public String name() {
    return this.name;
  }

  /** Returns all classes of this relation*/
  public Set<YagoClass> types() {
    return types;
  }

  /** Tells whether two relations are identical*/
  public boolean equals(Object obj) {
    return (obj == this);
  }

  /** Compares two relations by name*/
  public int compareTo(Relation o) {
    return (name.compareTo(o.name()));
  }

  /** Returns the hash of the name*/
  public int hashCode() {
    return name.hashCode();
  }

  public YagoClass domain() {
    return (domain);
  }

  public YagoClass range() {
    return (range);
  }

  public boolean isFunction() {
    return this.types.contains(YagoClasses.FUNCTION);
  }

  public boolean isTransitive() {
    return this.types.contains(YagoClasses.TRANSITIVERELATION);
  }

  public void addType(YagoClass type) {
    this.types.add(type);
  }
}
