package basics.datagraph;

import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.filehandlers.FileSet;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;

/**
 * Class Relations
 *
 * Provides a container for all Yago Relations
 * 
 * @author Edwin Lewis-Kelham, Fabian M. Suchanek
 */
public class Relations {

  // ----------------------------------------------------------
  // ---------------- Container -------------------------------
  // ----------------------------------------------------------

  /** Holds all created relations*/
  private static Map<String, Relation> values = new TreeMap<String, Relation>();

  /** All relations that apply to relations */
  public static Relation[] propertyList = new Relation[] {};

  /** adds a Relation to the container */
  public static void add(Relation relation) {
    if (values.containsKey(relation.name.toLowerCase())) {
      Announce.warning("Name '", relation.name.toLowerCase(), "' already exists in relation-container, new relation will override old one");
    } else {
      Announce.debug("Adding relation", relation.name(), " to relation-container");
    }
    values.put(relation.name.toLowerCase(), relation);
  }

  /** Maps a String to a Relation (or to null)*/
  public static Relation get(String name) {
    if (name == null) return null;
    return values.get(name.toLowerCase());
  }

  /**returns a collection containing all the relations*/
  public static Collection<Relation> values() {
    return (values.values());
  }

  // ----------------------------------------------------------
  // ---------------- Loading ---------------------------------
  // ----------------------------------------------------------

  private static final String RelationsPath = "data/Relations/Relation.xml";

  private static final String VirtualRelationsPath = "data/Relations/VirtualRelations.xml";

  public static void loadFromClass() {
    Announce.doing("Loading Relations from", "jar file");
    try {
      InputStream is = ClassLoader.getSystemResourceAsStream(RelationsPath);
      loadRelationsFromJar(is);
      is.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    try {
      InputStream is = ClassLoader.getSystemResourceAsStream(VirtualRelationsPath);
      loadRelationsFromJar(is);
      is.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    for (Relation yagoRelation : values()) {
      if (yagoRelation.superRelation == null) continue;
      Relation superrelation = get(yagoRelation.superRelation.name());
      if (superrelation == null) {
        Announce.warning("Undefined relation", yagoRelation.superRelation, "given as superrelation of", yagoRelation);
        continue;
      }
      yagoRelation.superRelation = superrelation;
    }
    Announce.done();
  }

  public static void loadRelationsFromJar(InputStream is) {
    SAXBuilder builder = new SAXBuilder();
    Announce.doing("Loading", "from relations inputstream");
    try {
      Document doc = builder.build(is);
      if (!doc.getRootElement().getName().equalsIgnoreCase("Relations")) {
        throw new Exception("The rootElement is " + doc.getRootElement().getName() + " and not Relations");
      }
      loadNames(doc);
      Announce.done();
    } catch (Exception e) {
      Announce.failed();
      Announce.warning(e);
      e.printStackTrace();
    }
  }

  /** Loads all YAGO classes*/
  public static boolean loadFrom(File yagoRelationFolder) {
    // Check the folder    
    Announce.doing("Loading Yago relations from folder", yagoRelationFolder);
    if (!yagoRelationFolder.exists()) {
      Announce.warning("Yago relations folder not found");
      Announce.failed();
      return (false);
    } else if (!yagoRelationFolder.isDirectory()) {
      Announce.warning("Yago relations folder is not a folder");
      Announce.failed();
      return (false);
    }
    // Load all files
    SAXBuilder builder = new SAXBuilder();
    for (File xmlFile : new FileSet(yagoRelationFolder, "*.xml")) {
      Announce.doing("Loading", xmlFile.getName());
      try {
        Document doc = builder.build(xmlFile);
        if (!doc.getRootElement().getName().equalsIgnoreCase("Relations")) {
          throw new Exception("The rootElement is " + doc.getRootElement().getName() + " and not Relations");
        }
        loadNames(doc);
        Announce.done();
      } catch (Exception e) {
        Announce.failed();
        Announce.warning(e);
        e.printStackTrace();
      }
    }
    // Canonicalize the super-relations
    for (Relation yagoRelation : values()) {
      if (yagoRelation.superRelation == null) continue;
      Relation superrelation = get(yagoRelation.superRelation.name());
      if (superrelation == null) {
        Announce.warning("Undefined relation", yagoRelation.superRelation, "given as superrelation of", yagoRelation);
        continue;
      }
      yagoRelation.superRelation = superrelation;
    }
    Announce.done();
    // Set the links
    setLinks();
    return (true);
  }

  /** Types of Yago Relations*/
  public enum RelationsType {
    RELATION, VIRTUALRELATION
  };

  @SuppressWarnings("unchecked")
  private static void loadNames(Document doc) {
    for (Element yagoRelation : (List<Element>) doc.getRootElement().getChildren()) {
      String name = yagoRelation.getAttributeValue("name");
      RelationsType type = null;
      try {
        type = RelationsType.valueOf(yagoRelation.getName().toUpperCase());
      } catch (Exception e) {
        Announce.warning("Unknown class type", type);
        continue;
      }
      YagoClass domain = YagoClasses.get(yagoRelation.getChildText("domain"));
      if (domain == null) Announce.warning("Undefined domain", yagoRelation.getChildText("domain"), " for Yago class", name);
      YagoClass range = YagoClasses.get(yagoRelation.getChildText("range"));
      if (range == null) Announce.warning("Undefined range", yagoRelation.getChildText("range"), " for Yago class", name);
      Relation superRelation = yagoRelation.getChildText("superRelation") == null || yagoRelation.getChildText("superRelation").length() == 0 ? null
          : new Relation(yagoRelation.getChildText("superRelation"), null, null);
      String sentence = yagoRelation.getChildText("sentence");
      List<YagoClass> types = new ArrayList<YagoClass>();
      Element typeList = yagoRelation.getChild("types");
      if (typeList != null) {
        for (Element superclass : (List<Element>) typeList.getChildren()) {
          types.add(new YagoClass(superclass.getText()));
        }
      }
      switch (type) {
        case RELATION:
          add(new Relation(name, sentence, domain, range, superRelation, types));
          break;
        case VIRTUALRELATION:
          String javaClassName = yagoRelation.getAttributeValue("class");
          try {
            Class javaClass = Class.forName(javaClassName);
            add((VirtualRelation) javaClass.getConstructors()[0].newInstance(name, domain, range, superRelation, types));
          } catch (Exception e) {
            Announce.warning("Could not instantiate Java class", javaClassName, "for Virtual relation", name);
          }
      }
    }
  }

  // ----------------------------------------------------------
  // ---------------- Links -----------------------------------
  // ----------------------------------------------------------

  public static Relation MEANS = null;

  public static Relation ISCALLED = null;

  public static Relation TYPE = null;

  public static Relation SUBCLASSOF = null;

  public static Relation DOMAIN = null;

  public static Relation RANGE = null;

  public static Relation SUBPROPERTYOF = null;

  public static Relation FAMILYNAME = null;

  public static Relation GIVENNAME = null;

  public static Relation DESCRIBES = null;

  public static Relation ESTABLISHEDONDATE = null;

  public static Relation HASWONPRIZE = null;

  public static Relation WRITTENINYEAR = null;

  public static Relation LOCATEDIN = null;

  public static Relation POLITICIANOF = null;

  public static Relation CONTEXT = null;

  public static Relation ISCITIZENOF = null;

  public static Relation MEREONYMY = null;

  public static Relation MEMBEROF = null;

  public static Relation SUBSTANCEOF = null;

  public static Relation PARTOF = null;

  public static Relation FOUNDIN = null;

  public static Relation USING = null;

  public static Relation DURING = null;

  public static Relation SINCE = null;

  public static Relation UNTIL = null;

  public static Relation INLANGUAGE = null;

  public static Relation HASVALUE = null;

  public static Relation INUNIT = null;

  public static Relation DISCOVERED = null;

  public static Relation DISCOVEREDONDATE = null;

  public static Relation BORNONDATE = null;

  public static Relation BORNIN = null;

  public static Relation ORIGIN = null;

  public static Relation DIEDONDATE = null;

  public static Relation DIEDIN = null;

  public static Relation NATIVENAME = null;

  public static Relation LEADER = null;

  public static Relation AREA = null;

  public static Relation POPULATION = null;

  public static Relation POPULATIONDENSITY = null;

  public static Relation UTCOFFSET = null;

  public static Relation WEBSITE = null;

  public static Relation ORDER = null;

  public static Relation PREDECESSOR = null;

  public static Relation SUCCESSOR = null;

  public static Relation MARRIEDTO = null;

  public static Relation AFFILIATEDTO = null;

  public static Relation INFLUENCED = null;

  public static Relation DIRECTED = null;

  public static Relation PRODUCED = null;

  public static Relation EDITED = null;

  public static Relation ACTEDIN = null;

  public static Relation PUBLISHEDONDATE = null;

  public static Relation DURATION = null;

  public static Relation PRODUCTIONLANGUAGE = null;

  public static Relation BUDGET = null;

  public static Relation IMDB = null;

  public static Relation PRODUCEDIN = null;

  public static Relation CHILD = null;

  public static Relation MOTTO = null;

  public static Relation OFFICIALLANGUAGE = null;

  public static Relation CAPITAL = null;

  public static Relation WATER = null;

  public static Relation GDP = null;

  public static Relation GDPN = null;

  public static Relation HDI = null;

  public static Relation GINI = null;

  public static Relation CURRENCY = null;

  public static Relation TIMEZONE = null;

  public static Relation TOPLEVELDOMAIN = null;

  public static Relation CALLINGCODE = null;

  public static Relation WROTE = null;

  public static Relation MADECOVER = null;

  public static Relation GENRE = null;

  public static Relation PUBLISHED = null;

  public static Relation PAGES = null;

  public static Relation ISBN = null;

  public static Relation JOINED = null;

  public static Relation LIVESIN = null;

  public static Relation HEIGHT = null;

  public static Relation WEIGHT = null;

  public static Relation SPOKENIN = null;

  public static Relation CREATED = null;

  public static Relation CREATEDONDATE = null;

  public static Relation INTERESTEDIN = null;

  public static Relation GROWTH = null;

  public static Relation INFLATION = null;

  public static Relation POVERTY = null;

  public static Relation LABOR = null;

  public static Relation UNEMPLOYMENT = null;

  public static Relation EXPORTVALUE = null;

  public static Relation EXPORTS = null;

  public static Relation IMPORTVALUE = null;

  public static Relation IMPORTS = null;

  public static Relation DEALSWITH = null;

  public static Relation REVENUE = null;

  public static Relation EXPENSES = null;

  public static Relation MILITARY = null;

  public static Relation NUMBEROFPEOPLE = null;

  public static Relation USESGDP = null;

  public static Relation WORKSAT = null;

  public static Relation ALMAMATER = null;

  public static Relation ACADEMICADVISOR = null;

  public static Relation HAPPENEDONDATE = null;

  public static Relation HAPPENEDIN = null;

  public static Relation PARTICIPATEDIN = null;

  public static Relation MUSICALROLE = null;

  public static Relation HASPRODUCT = null;

  // Virtual Relations
  public static VirtualRelation GREATERTHAN = null;

  public static VirtualRelation SMALLERTHAN = null;

  public static VirtualRelation BEFORE = null;

  public static VirtualRelation AFTER = null;

  public static VirtualRelation SAMEYEAR = null;

  public static VirtualRelation DISJOINT = null;

  public static VirtualRelation LIFETIME = null;

  public static VirtualRelation DIFFERENT = null;

  // From SOFIE evaluation, now obsolete
  public static Relation S_FOUNDEDONDATE = null;

  public static Relation S_BASEDIN = null;

  public static Relation S_ISHQ = null;

  public static Relation S_HQCITY = null;

  public static Relation S_OWNEDBY = null;

  public static Relation S_HASLANGUAGE = null;

  public static void setLinks() {
    MEANS = Relations.get("means");
    ISCALLED = Relations.get("isCalled");
    TYPE = Relations.get("type");
    SUBCLASSOF = Relations.get("subClassOf");
    DOMAIN = Relations.get("domain");
    RANGE = Relations.get("range");
    SUBPROPERTYOF = Relations.get("subPropertyOf");
    FAMILYNAME = Relations.get("familyNameOf");
    GIVENNAME = Relations.get("givenNameOf");
    DESCRIBES = Relations.get("describes");
    ESTABLISHEDONDATE = Relations.get("establishedOnDate");
    HASWONPRIZE = Relations.get("hasWonPrize");
    WRITTENINYEAR = Relations.get("writtenInYear");
    LOCATEDIN = Relations.get("locatedIn");
    POLITICIANOF = Relations.get("politicianOf");
    CONTEXT = Relations.get("context");
    ISCITIZENOF = Relations.get("isCitizenOf");
    MEREONYMY = Relations.get("isMereonymOf");
    MEMBEROF = Relations.get("isMemberOf");
    SUBSTANCEOF = Relations.get("isSubstanceOf");
    PARTOF = Relations.get("isPartOf");
    FOUNDIN = Relations.get("foundIn");
    USING = Relations.get("using");
    DURING = Relations.get("during");
    SINCE = Relations.get("since");
    UNTIL = Relations.get("until");
    INLANGUAGE = Relations.get("inLanguage");
    HASVALUE = Relations.get("hasValue");
    INUNIT = Relations.get("inUnit");
    DISCOVERED = Relations.get("discovered");
    DISCOVEREDONDATE = Relations.get("discoveredOnDate");

    BORNONDATE = Relations.get("bornOnDate");
    BORNIN = Relations.get("bornIn");
    ORIGIN = Relations.get("originatesFrom");
    DIEDONDATE = Relations.get("diedOnDate");
    DIEDIN = Relations.get("diedIn");
    NATIVENAME = Relations.get("isNativeNameOf");
    LEADER = Relations.get("isLeaderOf");
    AREA = Relations.get("hasArea");
    POPULATION = Relations.get("hasPopulation");
    POPULATIONDENSITY = Relations.get("hasPopulationDensity");
    UTCOFFSET = Relations.get("hasUTCOffset");
    WEBSITE = Relations.get("hasWebsite");
    ORDER = Relations.get("isNumber");
    PREDECESSOR = Relations.get("hasPredecessor");
    SUCCESSOR = Relations.get("hasSuccessor");
    MARRIEDTO = Relations.get("isMarriedTo");
    AFFILIATEDTO = Relations.get("isAffiliatedTo");
    INFLUENCED = Relations.get("influences");
    DIRECTED = Relations.get("directed");
    PRODUCED = Relations.get("produced");
    EDITED = Relations.get("edited");
    ACTEDIN = Relations.get("actedIn");
    PUBLISHEDONDATE = Relations.get("publishedOnDate");
    DURATION = Relations.get("hasDuration");
    PRODUCTIONLANGUAGE = Relations.get("hasProductionLanguage");
    BUDGET = Relations.get("hasBudget");
    IMDB = Relations.get("hasImdb");
    PRODUCEDIN = Relations.get("producedIn");
    CHILD = Relations.get("hasChild");
    MOTTO = Relations.get("hasMotto");
    OFFICIALLANGUAGE = Relations.get("hasOfficialLanguage");
    CAPITAL = Relations.get("hasCapital");
    WATER = Relations.get("hasWaterPart");
    GDP = Relations.get("hasGDPPPP");
    GDPN = Relations.get("hasNominalGDP");
    HDI = Relations.get("hasHDI");
    GINI = Relations.get("hasGini");
    CURRENCY = Relations.get("hasCurrency");
    TIMEZONE = Relations.get("inTimeZone");
    TOPLEVELDOMAIN = Relations.get("hasTLD");
    CALLINGCODE = Relations.get("hasCallingCode");
    WROTE = Relations.get("wrote");
    MADECOVER = Relations.get("madeCoverFor");
    GENRE = Relations.get("isOfGenre");
    PUBLISHED = Relations.get("published");
    PAGES = Relations.get("hasPages");
    ISBN = Relations.get("hasISBN");
    JOINED = Relations.get("joined");
    LIVESIN = Relations.get("livesIn");
    HEIGHT = Relations.get("hasHeight");
    WEIGHT = Relations.get("hasWeight");
    SPOKENIN = Relations.get("isSpokenIn");
    CREATED = Relations.get("created");
    CREATEDONDATE = Relations.get("createdOnDate");
    INTERESTEDIN = Relations.get("interestedIn");
    GROWTH = Relations.get("hasEconomicGrowth");
    INFLATION = Relations.get("hasInflation");
    POVERTY = Relations.get("hasPoverty");
    LABOR = Relations.get("hasLabor");
    UNEMPLOYMENT = Relations.get("hasUnemployment");
    EXPORTVALUE = Relations.get("hasExport");
    EXPORTS = Relations.get("exports");
    IMPORTVALUE = Relations.get("hasImport");
    IMPORTS = Relations.get("imports");
    DEALSWITH = Relations.get("dealsWith");
    REVENUE = Relations.get("hasRevenue");
    EXPENSES = Relations.get("hasExpenses");
    MILITARY = Relations.get("hasMilitary");
    NUMBEROFPEOPLE = Relations.get("hasNumberOfPeople");
    USESGDP = Relations.get("usesGDP");
    WORKSAT = Relations.get("worksAt");
    ALMAMATER = Relations.get("graduatedFrom");
    ACADEMICADVISOR = Relations.get("hasAcademicAdvisor");
    HAPPENEDONDATE = Relations.get("happenedOnDate");
    HAPPENEDIN = Relations.get("happenedIn");
    PARTICIPATEDIN = Relations.get("participatedIn");
    MUSICALROLE = Relations.get("musicalRole");
    HASPRODUCT = Relations.get("hasProduct");

    GREATERTHAN = (VirtualRelation) Relations.get(">");
    SMALLERTHAN = (VirtualRelation) Relations.get("<");
    BEFORE = (VirtualRelation) Relations.get("before");
    AFTER = (VirtualRelation) Relations.get("after");
    SAMEYEAR = (VirtualRelation) Relations.get("sameYear");
    DISJOINT = (VirtualRelation) Relations.get("disjoint");
    LIFETIME = (VirtualRelation) Relations.get("isLifeTime");
    DIFFERENT = (VirtualRelation) Relations.get("different");
    propertyList = new Relation[] { Relations.DOMAIN, Relations.RANGE, Relations.SUBPROPERTYOF, Relations.TYPE };
  }

  public static void main(String[] args) {
    YagoClasses.loadFrom(new File("c:\\Fabian\\EVE\\EVE2\\basics\\data\\YagoClasses"));
    loadFrom(new File("c:\\Fabian\\EVE\\EVE2\\basics\\data\\Relations"));
    for (Relation r : values()) {
      D.p(r.name(), r.range(), r.domain(), r.superRelation(), r.types());
    }
  }
}
