package basics.datagraph;

import java.util.ArrayList;
import java.util.List;

/**
 * This class is part of the YAGO extractors (http://mpii.de/yago). It is licensed under the 
 * Creative Commons Attribution-Noncommercial-Share Alike 3.0 Unported License,
 * (http://creativecommons.org/licenses/by-nc-sa/3.0/) 
 * by the YAGO team (http://mpii.de/yago). 
 * 
 * This class represents a YAGO class.<P>
 * 
 * Yago already imports tons of classes from WordNet. Some of them are of
 * particular importance for the Extractors/Generators and are thus represented
 * explicitly as YagoClasses. This class also defines YagoClasses that are not
 * in WordNet.
 * Relationjava converts these classes to the text tables.  <P>
 * 
 * Relation.java does not need to be adjusted if a YagoClass is added.
 * This class does need to be adjusted if a new version of WordNet is used! 
 * (currently 3.0)
 *
 * @author Fabian M. Suchanek
 * 
 * 
 * 
 * */

public class YagoClass implements Comparable<YagoClass> {

  /** Holds the name of the class */
  public String name;

  /** Holds the superclasses */
  protected List<YagoClass> superclasses;

  /** Constructor with full data*/
  public YagoClass(String name, List<YagoClass> superclasses) {
    this.name = name;
    this.superclasses = superclasses;
  }

  /** Constructor without superclasses*/
  public YagoClass(String name) {
    this(name,new ArrayList<YagoClass>(0));
  }

  /** Returns the direct superclasses */
  public List<YagoClass> directSuperclasses() {
    return superclasses;
  }

  /** Returns all superclasses */
  public List<YagoClass> allSuperclasses() {
    return allSuperclasses(new ArrayList<YagoClass>(4));
  }

  /** Returns all superclasses */
  protected List<YagoClass> allSuperclasses(List<YagoClass> l) {
    l.addAll(directSuperclasses());
    for (YagoClass c : directSuperclasses()) {
      c.allSuperclasses(l);
    }
    return (l);
  }

  /** Tells whether this class is a subclass of another (only within the YagoClasses)*/
  public boolean isSubClassOf(YagoClass s) {
    if (this == s) return (true);
    if (superclasses.size() == 0) return (false);
    for (YagoClass y : superclasses) {
      if (y.isSubClassOf(s)) return (true);
    }
    return (false);
  }

  /** Returns the name of this class */
  public String toString() {
    return (name);
  }

  /** Returns the name of this class */
  public String name() {
    return (name);
  }

  @Override
  public int hashCode() {
    return name.hashCode();
  }

  public int compareTo(YagoClass o) {
    return this.name.compareTo(o.name);
  }

  /** Adds a super class */
  public void addSuperClass(YagoClass yagoClass) {
    if (yagoClass != null) {
      superclasses.add(yagoClass);
    }
  }

}
