package basics.datagraph;

import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.filehandlers.FileSet;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;

/**
 * Class YagoClasses
 *
 * Provides a container for all Yago Classes
 * 
 * @author Edwin Lewis-Kelham, Fabian M. Suchanek
 */
public class YagoClasses {

  // ----------------------------------------------------------
  // ---------------- Container -------------------------------
  // ----------------------------------------------------------

  /** Holds all Yago Classes */
  protected static Map<String, YagoClass> values = new TreeMap<String, YagoClass>();

  /** adds a YagoClass to the container */
  public static void add(YagoClass yagoClass) {
    String name = yagoClass.name();
    if (yagoClass instanceof WordNetClass) {
      values.put(yagoClass.name().toLowerCase(), yagoClass);
      name = ((WordNetClass) yagoClass).mainConcept();
    }
    name = name.toLowerCase();
    if (values.containsKey(name)) {
      Announce.warning("Name '", name, "' already exists in yagoclass-container, new yagoclass will override old one");
    } else {
      Announce.debug("Adding yagoclass", yagoClass, "with name", name, "to yagoclass-container");
    }
    values.put(name, yagoClass);
  }

  /** Maps a String to a YagoClass (or to null)*/
  public static YagoClass get(String name) {
    if (name == null) return null;
    return values.get(name.toLowerCase());
  }

  /**returns a collection containing all the yagoClasses*/
  public static Collection<YagoClass> values() {
    return (values.values());
  }

  // ----------------------------------------------------------
  // ---------------- Loading ---------------------------------
  // ----------------------------------------------------------
  private static final String LiteralYagoClassesPath = "data/YagoClasses/LiteralClasses.xml";

  private static final String OtherYagoClassesPath = "data/YagoClasses/OtherClasses.xml";

  private static final String WordNetYagoClassesPath = "data/YagoClasses/WordNetClasses.xml";

  public static void loadFromClass() {
    Announce.doing("Loading YagoClasses from", "jar file");
    try {
      InputStream is = ClassLoader.getSystemResourceAsStream(LiteralYagoClassesPath);
      loadYagoClassFromJar(is);
      is.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    try {
      InputStream is = ClassLoader.getSystemResourceAsStream(OtherYagoClassesPath);
      loadYagoClassFromJar(is);
      is.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    try {
      InputStream is = ClassLoader.getSystemResourceAsStream(WordNetYagoClassesPath);
      loadYagoClassFromJar(is);
      is.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    Announce.done();
  }
  
  private static void loadYagoClassFromJar(InputStream is) {
    SAXBuilder builder = new SAXBuilder();
    Announce.doing("Loading", "from yagoclasses inputstream");
    try {
      Document doc = builder.build(is);
      if (!doc.getRootElement().getName().equalsIgnoreCase("YagoClasses")) {
        throw new Exception("The rootElement is " + doc.getRootElement().getName() + " and not YagoClasses");
      }
      loadNames(doc);
      Announce.done();
    } catch (Exception e) {
      Announce.failed();
      Announce.warning(e);
      e.printStackTrace();
    }
  }

  /** Loads all YAGO classes*/
  public static boolean loadFrom(File yagoClassFolder) {
    // Check the folder    
    Announce.doing("Loading YagoClasses from folder", yagoClassFolder);
    if (!yagoClassFolder.exists()) {
      Announce.warning("YagoClass folder not found");
      Announce.failed();
      return (false);
    } else if (!yagoClassFolder.isDirectory()) {
      Announce.warning("YagoClass folder is not a folder");
      Announce.failed();
      return (false);
    }
    // Load all files
    SAXBuilder builder = new SAXBuilder();
    for (File xmlFile : new FileSet(yagoClassFolder, "*.xml")) {
      Announce.doing("Loading", xmlFile.getName());
      try {
        Document doc = builder.build(xmlFile);
        if (!doc.getRootElement().getName().equalsIgnoreCase("YagoClasses")) {
          throw new Exception("The rootElement is " + doc.getRootElement().getName() + " and not YagoClasses");
        }
        loadNames(doc);
        Announce.done();
      } catch (Exception e) {
        Announce.failed();
        Announce.warning(e);
        e.printStackTrace();
      }
    }
    // Canonicalize the super-classes
    for (YagoClass yagoClass : values()) {
      for (int i = 0; i < yagoClass.superclasses.size(); i++) {
        YagoClass superclass = get(yagoClass.superclasses.get(i).name());
        if (superclass == null) {
          Announce.warning("Undefined class", yagoClass.superclasses.get(i).name(), "given as superclass of", yagoClass);
          continue;
        }
        yagoClass.superclasses.set(i, superclass);
      }
    }
    Announce.done();
    // Set the links
    setLinks();
    return (true);
  }

  /** Types of Yago Classes*/
  public enum YagoClassType {
    WORDNETCLASS, LITERALYAGOCLASS, YAGOCLASS
  };

  @SuppressWarnings("unchecked")
  private static void loadNames(Document doc) {
    for (Element yagoClass : (List<Element>) doc.getRootElement().getChildren()) {
      String name = yagoClass.getAttributeValue("name");
      YagoClassType type = null;
      try {
        type = YagoClassType.valueOf(yagoClass.getName().toUpperCase());
      } catch (Exception e) {
        Announce.warning("Unknown class type", type);
        continue;
      }
      List<YagoClass> superclasses = new ArrayList<YagoClass>();
      Element superclassesList = yagoClass.getChild("superclasses");
      if (superclassesList != null) {
        for (Element superclass : (List<Element>) superclassesList.getChildren()) {
          superclasses.add(new YagoClass(superclass.getText()));
        }
      }
      switch (type) {
        case WORDNETCLASS:
          try {
            int id = Integer.parseInt(yagoClass.getAttributeValue("id"));
            YagoClasses.add(new WordNetClass(name, id));
          } catch (Exception e) {
            Announce.warning("Could not parse id", yagoClass.getAttributeValue("id"), "of WordNetClass", name);
          }
          break;
        case LITERALYAGOCLASS:
          String pattern = yagoClass.getChildText("typeCheckPattern");
          if (pattern != null && pattern.length() == 0) pattern = null;
          YagoClasses.add(new LiteralYagoClass(name, superclasses, pattern));
          break;
        case YAGOCLASS:
          YagoClasses.add(new YagoClass(name, superclasses));
      }
    }
  }

  // ----------------------------------------------------------
  // ---------------- Links -----------------------------------
  // ----------------------------------------------------------

  public static YagoClass USSTATE = null;

  public static WordNetClass POLITICIAN = null; //TODO

  public static WordNetClass ABSTRACT = null;

  public static WordNetClass PHYSICAL = null;

  public static WordNetClass ARTIFACT = null;

  public static WordNetClass PERSON = null;

  public static WordNetClass LEGALACTOR = null;

  public static WordNetClass LEGALACTORGEO = null;

  public static WordNetClass BOOK = null;

  public static WordNetClass ALBUM = null;

  public static WordNetClass UNIT = null;

  public static WordNetClass COMPANY = null;

  public static WordNetClass ORGANIZATION = null;

  public static WordNetClass CITY = null;

  public static WordNetClass GEOPOLITICAL = null;

  public static WordNetClass COUNTRY = null;

  public static WordNetClass ENTITY = null;

  public static WordNetClass MOVIE = null;

  public static WordNetClass ACTOR = null;

  public static WordNetClass DATABASE = null;

  public static WordNetClass COMPUTERSYSTEM = null;

  public static WordNetClass COMPUTERSCIENTIST = null;

  public static WordNetClass PRIZE = null;

  public static WordNetClass UNIVERSITY = null;

  public static WordNetClass LANGUAGE = null;

  public static YagoClass CLASS = null;

  public static YagoClass RELATION = null;

  public static YagoClass TRANSITIVERELATION = null;

  public static YagoClass SYMMETRICRELATION = null;

  public static YagoClass FUNCTION = null;

  public static YagoClass FACT = null;

  public static WordNetClass ARTIST = null;

  public static WordNetClass SINGER = null;

  public static WordNetClass SONG = null;

  public static LiteralYagoClass MUSIC_BRAINS_ID = null;

  public static LiteralYagoClass PROPORTION = null;

  public static LiteralYagoClass WORDINCONTEXT = null;

  public static LiteralYagoClass LITERAL = null;

  public static LiteralYagoClass BOOLEAN = null;

  public static LiteralYagoClass NUMBER = null;

  public static LiteralYagoClass RATIONAL = null;

  public static LiteralYagoClass INTEGER = null;

  public static LiteralYagoClass NONNEGATIVEINTEGER = null;

  public static LiteralYagoClass STRING = null;

  public static LiteralYagoClass WORD = null;

  public static LiteralYagoClass URL = null;

  public static LiteralYagoClass TLD = null;

  public static LiteralYagoClass CHAR = null;

  public static LiteralYagoClass TIMEINTERVAL = null;

  public static LiteralYagoClass TIMEPOINT = null;

  public static LiteralYagoClass DATE = null;

  public static LiteralYagoClass YEAR = null;

  public static LiteralYagoClass QUANTITY = null;

  public static LiteralYagoClass LENGTH = null;

  public static LiteralYagoClass AREA = null;

  public static LiteralYagoClass WEIGHT = null;

  public static LiteralYagoClass DENSITYPERAREA = null;

  public static LiteralYagoClass DURATION = null;

  public static LiteralYagoClass MONETARYVALUE = null;

  public static LiteralYagoClass IDENTIFIER = null;

  public static LiteralYagoClass CALLINGCODE = null;

  public static LiteralYagoClass ISBN = null;

  protected static void setLinks() {
    USSTATE = YagoClasses.get("wikicategory_States_of_the_United_States");
    POLITICIAN = (WordNetClass) YagoClasses.get("senator"); // TODO
    ABSTRACT = (WordNetClass) YagoClasses.get("abstract_entity");
    PHYSICAL = (WordNetClass) YagoClasses.get("physical_entity");
    ARTIFACT = (WordNetClass) YagoClasses.get("artifact");
    PERSON = (WordNetClass) YagoClasses.get("person");
    LEGALACTOR = (WordNetClass) YagoClasses.get("yagoActor");
    LEGALACTORGEO = (WordNetClass) YagoClasses.get("yagoActorGeo");
    BOOK = (WordNetClass) YagoClasses.get("novel");
    ALBUM = (WordNetClass) YagoClasses.get("medium");
    UNIT = (WordNetClass) YagoClasses.get("unit_of_measurement");
    COMPANY = (WordNetClass) YagoClasses.get("company");
    ORGANIZATION = (WordNetClass) YagoClasses.get("organization");
    CITY = (WordNetClass) YagoClasses.get("city");
    GEOPOLITICAL = (WordNetClass) YagoClasses.get("location");
    COUNTRY = (WordNetClass) YagoClasses.get("country");
    ENTITY = (WordNetClass) YagoClasses.get("entity");
    MOVIE = (WordNetClass) YagoClasses.get("movie");
    ACTOR = (WordNetClass) YagoClasses.get("actor");
    DATABASE = (WordNetClass) YagoClasses.get("database");
    COMPUTERSYSTEM = (WordNetClass) YagoClasses.get("program");
    COMPUTERSCIENTIST = (WordNetClass) YagoClasses.get("computer_scientist");
    PRIZE = (WordNetClass) YagoClasses.get("award");
    UNIVERSITY = (WordNetClass) YagoClasses.get("university");
    LANGUAGE = (WordNetClass) YagoClasses.get("language");
    CLASS = YagoClasses.get("yagoClass");
    RELATION = YagoClasses.get("yagoRelation");
    TRANSITIVERELATION = YagoClasses.get("yagoTransitiveRelation");
    SYMMETRICRELATION = YagoClasses.get("yagoSymmetricRelation");
    FUNCTION = YagoClasses.get("yagoFunction");
    FACT = YagoClasses.get("yagoFact");
    ARTIST = (WordNetClass) YagoClasses.get("performer");
    SINGER = (WordNetClass) YagoClasses.get("singer");
    SONG = (WordNetClass) YagoClasses.get("song");

    MUSIC_BRAINS_ID = (LiteralYagoClass) YagoClasses.get("musicBrainsId");
    PROPORTION = (LiteralYagoClass) YagoClasses.get("yagoProportion");
    WORDINCONTEXT = (LiteralYagoClass) YagoClasses.get("yagoWordInContext");
    LITERAL = (LiteralYagoClass) YagoClasses.get("yagoLiteral");
    BOOLEAN = (LiteralYagoClass) YagoClasses.get("yagoBoolean");
    NUMBER = (LiteralYagoClass) YagoClasses.get("yagoNumber");
    RATIONAL = (LiteralYagoClass) YagoClasses.get("yagoRationalNumber");
    INTEGER = (LiteralYagoClass) YagoClasses.get("yagoInteger");
    NONNEGATIVEINTEGER = (LiteralYagoClass) YagoClasses.get("yagoNonNegativeInteger");
    STRING = (LiteralYagoClass) YagoClasses.get("yagoString");
    WORD = (LiteralYagoClass) YagoClasses.get("yagoWord");
    URL = (LiteralYagoClass) YagoClasses.get("yagoURL");
    TLD = (LiteralYagoClass) YagoClasses.get("yagoTLD");
    CHAR = (LiteralYagoClass) YagoClasses.get("yagoChar");
    TIMEINTERVAL = (LiteralYagoClass) YagoClasses.get("yagoTimeInterval");
    TIMEPOINT = (LiteralYagoClass) YagoClasses.get("yagoTimePoint");
    DATE = (LiteralYagoClass) YagoClasses.get("yagoDate");
    YEAR = (LiteralYagoClass) YagoClasses.get("yagoYear");
    QUANTITY = (LiteralYagoClass) YagoClasses.get("yagoQuantity");
    LENGTH = (LiteralYagoClass) YagoClasses.get("yagoLength");
    AREA = (LiteralYagoClass) YagoClasses.get("yagoArea");
    WEIGHT = (LiteralYagoClass) YagoClasses.get("yagoWeight");
    DENSITYPERAREA = (LiteralYagoClass) YagoClasses.get("yagoDensityPerArea");
    DURATION = (LiteralYagoClass) YagoClasses.get("yagoDuration");
    MONETARYVALUE = (LiteralYagoClass) YagoClasses.get("yagoMonetaryValue");
    IDENTIFIER = (LiteralYagoClass) YagoClasses.get("yagoIdentifier");
    CALLINGCODE = (LiteralYagoClass) YagoClasses.get("yagoCallingCode");
    ISBN = (LiteralYagoClass) YagoClasses.get("yagoISBN");
  }

  // ----------------------------------------------------------
  // ---------------- Test ------------------------------------
  // ----------------------------------------------------------

  public static void main(String[] args) {
    loadFrom(new File("c:/fabian/eve/eve2/basics/data/yagoclasses"));
    for (YagoClass y : values()) {
      D.p(y.name, y.superclasses);
      if (y instanceof LiteralYagoClass) D.p(((LiteralYagoClass) y).typeCheckPattern);
      if (y instanceof WordNetClass) D.p(((WordNetClass) y).wordNetId());
    }
  }
}
