package converters;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.Writer;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.UTF8Writer;
import javatools.parsers.Char;
import basics.datagraph.DataGraphManager;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.YagoClasses;

/** 
 * This class is part of the YAGO converters (http://mpii.de/yago). 
 * It is licensed under the Creative Commons Attribution License 
 * (http://creativecommons.org/licenses/by/3.0) 
 * by the YAGO team (http://mpii.de/yago).
 * 
 * This class converts YAGO to the DB format required by Alchemy
 * 
 * @author Fabian M. Suchanek
 */
public class AlchConverter {

  /**
   * the yago folder
   */
  public static File yagoFolder;

  /**
   * target output folder
   */
  public static File outputFolder;

  /**
   * set to true to run as test only top 100 of each file 
   */
  public static boolean test = false;

  /**
   * minimises the entity name
   */
  public static String mlnNameFor(Object entity) {
    String name = entity.toString();
    if (name.startsWith("wikicategory_")) name = "cat_" + name.substring(13);
    if (name.startsWith("wordnet_")) name = "wn_" + name.substring(8);
    name = Char.upCaseFirst(Char.encodeHex(name));
    if (Character.isDigit(name.charAt(0))) name = "Y" + name;
    return (name);
  }

  /**
   * 
   * creates the entity list
   *    @param out 
   */
  public static void entityList(Writer out) throws Exception {
    Announce.doing("Creating entity list");
    out.write("entity = {");
    for (File f : new File(yagoFolder, "entities").listFiles()) {
      int counter = 0;
      FileLines lines = new FileLines(f, "Parsing " + f.getName());
      for (String line : lines) {
        line = line.split("\t")[0];
        out.write(mlnNameFor(line) + ",");
        if (test && counter++ > 100) break;
      }
      lines.close();
    }
    for (Relation r : Relations.values()) {
      if (r.name().startsWith("_")) continue;
      out.write(mlnNameFor(r) + ",");
    }
    D.writeln(out, "YagoDummy}");
    Announce.done();
  }

  /**
   * creates the fact list
   * @param outputFolder file
   * @throws Exception
   */
  public static void facts(File outputFolder) throws Exception {
    Writer out = new BufferedWriter(new FileWriter(new File(outputFolder, "yago.db")));
    for (File relationName : new File(yagoFolder, "facts").listFiles()) {
      Relation relation = Relations.get(relationName.getName());
      if (relation == null) continue;
      if (relation.domain().isSubClassOf(YagoClasses.LITERAL)) continue;
      if (relation.range().isSubClassOf(YagoClasses.LITERAL)) continue;
      for (File file : relationName.listFiles()) {
        int counter = 0;
        FileLines lines = new FileLines(file, "Parsing " + file.getName());
        for (String line : lines) {
          String[] split = line.split("\t");
          D.writeln(out, "fact(" + mlnNameFor(split[1]) + ", " + mlnNameFor(relation) + ", " + mlnNameFor(split[2]) + ")");
          if (test && counter++ > 100) break;
        }
        lines.close();
      }
    }
    out.close();
  }

  /**
   * main method requires no args but a yago ini path can be provided
   * @param args
   * @throws Exception
   */
  public static void main(String[] args) throws Exception {
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFile(f);
    } else {
      requestInformation();
    }
  }

  /**
   * reads the parameters from the ini file
   * @param initFile
   * @throws Exception
   */
  public static void fromIniFile(File initFile) throws Exception {
    Parameters.init(initFile);
    DataGraphManager.init();
    Parameters.ensureParameters("yagoFolder - the YAGO data folder", "alchFolder - the folder where the Alchemy data shall go");
    yagoFolder = Parameters.getFile("yagoFolder");
    outputFolder = Parameters.getFile("alchFolder");
    Announce.doing("Producing Alchemy version of YAGO");
    Announce.doing("Creating Alchemy folder");
    outputFolder.mkdir();
    Announce.done();
    convert();
  }

  /** request information instead of using yago.ini file*/
  public static void requestInformation() throws Exception {
    while (yagoFolder == null) {
      D.println("Please enter a yago Folder directory containing the yago facts and entities (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      yagoFolder = new File(path);
      if (!yagoFolder.exists()) {
        D.println("yago Folder does not exist, please try again (quit to end)");
        yagoFolder = null;
      } else {
        if (!new File(yagoFolder, "facts").exists()) {
          D.println(path + " does not have a sub-folder 'facts', please try again (quit to end)");
          yagoFolder = null;
        }
        if (!new File(yagoFolder, "entities").exists()) {
          D.println(path + " does not have a sub-folder 'entities', please try again (quit to end)");
          yagoFolder = null;
        }
      }
    }
    while (outputFolder == null) {
      D.println("Please enter a target directory (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      outputFolder = new File(path);
      if (!outputFolder.exists()) {
        if (!outputFolder.mkdir()) {
          outputFolder = null;
        }
      }
    }
    boolean ok = false;
    while (!ok) {
      D.println("");
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
    }
    convert();
  }

  /**
   * once all parameters are provided the converter can be started
   * @throws Exception
   */
  private static void convert() throws Exception {
    D.println("Starting conversion to Alchemy");
    Announce.doing("Writing rules");
    Writer out = new UTF8Writer(new File(outputFolder, "yago.mln"));
    entityList(out);
    D.writeln(out, "\nfact(entity, entity, entity)\n");
    D.writeln(out, "fact(x,Type,y) ^ fact(y,SubclassOf,z) => fact(x,Type,z).");
    D.writeln(out, "fact(r,Domain,d) ^ fact(x,r,y) => fact(x,Type,d).");
    D.writeln(out, "fact(r,Range,d) ^ fact(x,r,y) => fact(y,Type,d).");
    D.writeln(out, "100 fact(r,Type,YagoFunction) ^ fact(x,r,y) ^ fact(x,r,y2) => y=y2");
    out.close();
    Announce.done();
    Announce.doing("Writing facts");
    facts(outputFolder);
    out.close();
    Announce.done();
    Announce.done();
  }

}
