package converters;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Collection;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.FileSet;
import javatools.filehandlers.UTF8Reader;
import javatools.filehandlers.UTF8Writer;
import javatools.parsers.Char;
import basics.datagraph.DataGraphManager;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.YagoClasses;

/** 
 * This class is part of the YAGO converters (http://mpii.de/yago). 
 * It is licensed under the Creative Commons Attribution License 
 * (http://creativecommons.org/licenses/by/3.0) 
 * by the YAGO team (http://mpii.de/yago).
 * 
 * This class produces the N3 triples for the relations
 *   MEANS, TYPE, SUBCLASSOF
 * in order to support the import of these facts to DBpedia.
 * 
 * This code is intended as a boilerplate for the DBpedia team.
 * The YAGO team does not actively maintain this code
 * and does not guarantee that the code produces the desired output for DBpedia. 
 * 
 * @author Fabian M. Suchanek
 */
public class DBpediaExport {

  /** holds the YAGO folder*/
  public static File yagoFolder;

  /** holds the folder for the DBpedia output*/
  public static File dbpediaFolder;

  /** Represents a relation that is to be converted*/
  public static class ConvertRelation {

    /** The Relations (such as Relation.SUBCLASSOF)*/
    protected Relation yagoRelation;

    /** TRUE if N3 requires the arguments to be swapped (like a MEANS b --> b label a)*/
    protected boolean inverted;

    /** Name of the relation in N3 (e.g. rdfs:label)*/
    protected String n3Name;

    /** Yago data files that shall be converted (NULL for all)*/
    protected Collection<String> files;

    /** Tells whether the second N3 argument is a string*/
    protected boolean secondIsString;

    /**Constructor*/
    public ConvertRelation(Relation yagoRelation, boolean inverted, String name, String... files) {
      this.yagoRelation = yagoRelation;
      this.inverted = inverted;
      n3Name = name;
      this.files = files.length == 0 ? null : Arrays.asList(files);
      secondIsString = inverted ? yagoRelation.domain().isSubClassOf(YagoClasses.STRING) : yagoRelation.range().isSubClassOf(YagoClasses.STRING);
    }

    /** TRUE if the second argument of the N3 triple must be a string*/
    public boolean secondIsString() {
      return (secondIsString);
    }

    /** Returns the yago data file names that shall be converted*/
    public Collection<String> fileNames() {
      if (files == null) return (Arrays.asList(FileSet.file(yagoFolder, "facts", yagoRelation.name()).list()));
      return (files);
    }

    /** Returns the File object with path for a yago data file name*/
    public File file(String fileName) {
      return (FileSet.file(yagoFolder, "facts", yagoRelation.name(), fileName));
    }

    /** Tells whether the arguments in N3 and YAGO are swapped*/
    public boolean isInverted() {
      return inverted;
    }

    /** Returns the name of this relation in N3 */
    public String n3Name() {
      return n3Name;
    }

    /** Returns the YAGO relation*/
    public Relation yagoRelation() {
      return yagoRelation;
    }

    @Override
    public String toString() {
      return yagoRelation + " -> " + n3Name;
    }
  }

  /** Relations that are to be converted */
  public static Collection<ConvertRelation> convertRelations = null;

  private static void prepare() {
    convertRelations = Arrays.asList(new ConvertRelation(Relations.MEANS, true, "rdfs:label", "ConceptLinker.txt", "WordNetLinks.txt"), new ConvertRelation(Relations.SUBCLASSOF, false, "rdfs:subClassOf"), new ConvertRelation(Relations.TYPE, false,
        "rdf:type", "ArticleExtractor.txt", "CheckedFactExtractor.txt"));
  }

  /** Creates a DBpedia identifier for a YAGO identifier.
   * yagoPrefix is TRUE for the prefix "yago:" and FALSE for the prefix "dbpedia:" */
  public static String dbpediaNameFor(String yagoName, boolean yagoPrefix, boolean asString) {
    if (asString) return ('"' + Char.encodeBackslash(Char.cutLast(yagoName.substring(1))) + '"');
    yagoName = Char.encodeURIPathComponent(yagoName);
    if (!yagoPrefix) return ("<http://dbpedia.org/resource/" + yagoName + ">");
    if (yagoName.startsWith("wordnet_")) yagoName = yagoName.substring(8);
    if (yagoName.startsWith("wikicategory_")) yagoName = yagoName.substring(13);
    String[] split = yagoName.split("_");
    StringBuilder result = new StringBuilder("<http://dbpedia.org/class/yago/");
    for (String s : split)
      result.append(Char.upCaseFirst(s));
    result.append('>');
    return (result.toString());
  }

  /** Produces the N3 output for DBpedia */
  public static void convertFacts() throws IOException {
    D.println("Starting DBpedia export");
    Announce.doing("Exporting");
    Writer out = new UTF8Writer(new File(dbpediaFolder, "DBpediaYAGOexport.n3"));
    D.writeln(out, "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .");
    D.writeln(out, "@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .");
    for (ConvertRelation convertRelation : convertRelations) {
      boolean useDBpediaPrefix = convertRelation.yagoRelation() != Relations.TYPE;
      Announce.doing("Converting", convertRelation);
      for (String yagoFileName : convertRelation.fileNames()) {
        //int counter=0;
        for (String l : new FileLines(new UTF8Reader(convertRelation.file(yagoFileName), "Parsing " + yagoFileName))) {
          //if(counter++>100) break;
          String[] split = l.split("\t");
          if (split.length != 4) continue;
          String arg1 = dbpediaNameFor(split[convertRelation.isInverted() ? 2 : 1], useDBpediaPrefix, false);
          String arg2 = dbpediaNameFor(split[convertRelation.isInverted() ? 1 : 2], true, convertRelation.secondIsString());
          out.write(arg1 + " " + convertRelation.n3Name() + " " + arg2 + " .\n");
        }
      }
    }
    out.close();
    Announce.done();
  }

  /**
   * reads the parameters from the ini file
   * @param initFile
   * @throws Exception
   */
  public static void fromIniFile(File initFile) throws Exception {
    Parameters.init(initFile);
    DataGraphManager.init();
    prepare();
    Parameters.ensureParameters("yagoFolder - the YAGO data folder", "dbpediaFolder - the folder where the DBpedia output shall go");
    yagoFolder = Parameters.getFile("yagoFolder");
    dbpediaFolder = Parameters.getFile("dbpediaFolder");
    Announce.doing("Creating DBpedia output folder");
    dbpediaFolder.mkdir();
    Announce.done();
    convertFacts();
    Announce.done();
  }

  /**
   * main method for starting a DBpedia export. ini file does not need to be provided.
   */
  public static void main(String[] args) throws Exception {
    Announce.doing("Exporting YAGO data for DBpedia");
    Announce.message("\n  This program is a boilerplate program for the DBpedia team.");
    Announce.message("The YAGO team does not actively maintain this program");
    Announce.message("and does not guarantee that it produces the desired output for DBpedia.\n");
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFile(f);
    } else {
      requestInformation();
    }
  }

  /** request information instead of using yago.ini file*/
  public static void requestInformation() throws Exception {
    while (yagoFolder == null) {
      D.println("Please enter a yago Folder directory containing the yago facts and entities (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      yagoFolder = new File(path);
      if (!yagoFolder.exists()) {
        D.println("yago Folder does not exist, please try again (quit to end)");
        yagoFolder = null;
      } else {
        if (!new File(yagoFolder, "facts").exists()) {
          D.println(path + " does not have a sub-folder 'facts', please try again (quit to end)");
          yagoFolder = null;
        }
        if (!new File(yagoFolder, "entities").exists()) {
          D.println(path + " does not have a sub-folder 'entities', please try again (quit to end)");
          yagoFolder = null;
        }
      }
    }
    while (dbpediaFolder == null) {
      D.println("Please enter a target directory (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      dbpediaFolder = new File(path);
      if (!dbpediaFolder.exists()) {
        if (!dbpediaFolder.mkdir()) {
          dbpediaFolder = null;
        }
      }
    }
    boolean ok = false;
    while (!ok) {
      D.println("");
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
      if (ok) {
        prepare();
      }
    }
    convertFacts();
  }
}
