package converters;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.Writer;

import basics.datagraph.DataGraphManager;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.UTF8Reader;
import javatools.parsers.Char;

/**
 * Class LinkingOpenData
 * 
 * This class is part of the YAGO converters (http://mpii.de/yago). 
 * It is licensed under the Creative Commons Attribution License 
 * (http://creativecommons.org/licenses/by/3.0) 
 * by the YAGO team (http://mpii.de/yago).
 * 
 * This class produces the links between DBpedia and YAGO in the spirit of
 * the Linking Open data project. More precisely, it produces
 * 
 * <http://dbpedia.org/class/yago/class-name>
 *     owl:equivalentClass
 * <http://mpii.de/yago/resource/class-name>
 *
 *        and
 *   
 * <http://dbpedia.org/resource/individual-name>
 *     owl:sameAs
 * <http://mpii.de/yago/resource/individual-name>
 * 
 * @author Fabian M. Suchanek
 */
public class DBpediaLink {

  /**
   * yago folder
   */
  private static File yagoFolder;

  /**
   * target folder
   */
  private static File dbpediaFolder;

  /**
   * main method can either use an ini file or all parameters will be requested
   * @param args
   * @throws Exception
   */
  public static void main(String[] args) throws Exception {
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFile(f);
    } else {
      requestInformation();
    }
  }

  /** request information instead of using yago.ini file*/
  public static void requestInformation() throws Exception {
    while (yagoFolder == null) {
      D.println("Please enter a yago Folder directory containing the yago facts and entities (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      yagoFolder = new File(path);
      if (!yagoFolder.exists()) {
        D.println("yago Folder does not exist, please try again (quit to end)");
        yagoFolder = null;
      } else {
        if (!new File(yagoFolder, "facts").exists()) {
          D.println(path + " does not have a sub-folder 'facts', please try again (quit to end)");
          yagoFolder = null;
        }
        if (!new File(yagoFolder, "entities").exists()) {
          D.println(path + " does not have a sub-folder 'entities', please try again (quit to end)");
          yagoFolder = null;
        }
      }
    }
    while (dbpediaFolder == null) {
      D.println("Please enter a target directory (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      dbpediaFolder = new File(path);
      if (!dbpediaFolder.exists()) {
        if (!dbpediaFolder.mkdir()) {
          dbpediaFolder = null;
        }
      }
    }
    boolean ok = false;
    while (!ok) {
      D.println("");
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
    }
    convert();
  }

  /**
   * reads the parameters from an ini file
   * @param initFile
   * @throws Exception
   */
  public static void fromIniFile(File initFile) throws Exception {
    Parameters.init(initFile);
    DataGraphManager.init();
    Parameters.ensureParameters("yagoFolder - the YAGO data folder", "dbpediaFolder - the output target folder");
    yagoFolder = Parameters.getFile("yagoFolder");
    dbpediaFolder = Parameters.getFile("dbpediaFolder");
    Announce.doing("Creating DBpedia output folder");
    dbpediaFolder.mkdir();
    Announce.done();
    convert();
  }

  /**
   * starts to link DBpedia and YAGO
   * @throws Exception
   */
  public static void convert() throws Exception {
    Announce.doing("Strating to link DBpedia and YAGO");
    Writer out = new BufferedWriter(new FileWriter(new File(dbpediaFolder, "DBpediaYAGOlink.n3")));
    D.writeln(out, "@prefix owl:  <http://www.w3.org/2002/07/owl#> .");
    for (File entityFile : new File(yagoFolder, "entities").listFiles()) {
      for (String line : new FileLines(new UTF8Reader(entityFile, "Parsing " + entityFile.getName()))) {
        String name = line.substring(0, line.indexOf('\t'));
        name = Char.encodeURIPathComponent(name);
        if (line.endsWith("false")) {
          D.writeln(out, "<http://mpii.de/yago/resource/" + name + "> owl:sameAs <http://dbpedia.org/resource/" + name + "> .");
        } else {
          D.writeln(out, "<http://mpii.de/yago/resource/" + name + "> owl:sameAs <http://dbpedia.org/class/yago/" + name + "> .");
        }
      }
    }
    out.close();
    Announce.done();
  }
}
