package converters;

import java.io.File;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.database.Database;
import javatools.database.MySQLDatabase;
import javatools.database.OracleDatabase;
import javatools.database.PostgresDatabase;
import javatools.datatypes.PeekIterator;
import basics.configuration.DBConfig;
import basics.datagraph.DataGraphManager;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.VirtualRelation;
import basics.datagraph.YagoClasses;

/** 
 * This class is part of the YAGO converters (http://mpii.de/yago). 
 * It is licensed under the Creative Commons Attribution License 
 * (http://creativecommons.org/licenses/by/3.0) 
 * by the YAGO team (http://mpii.de/yago).
 *
 * If you use this class for scientific purposes, please cite our paper
 *     Fabian M. Suchanek, Gjergji Kasneci and Gerhard Weikum
 *     "Yago - A Core of Semantic Knowledge" (WWW 2007)
 * 
 * This class implements a simple query processor for reification graphs as described in the Yago journal paper.
 * 
 * @author Fabian M. Suchanek
 */
public class QueryProcessor {

  //----------------------------------------------------
  //           General classes and helper routines
  //----------------------------------------------------

  /** maximal elements of an SQL query to retrieve*/
  public final int MAXSQLRES = 1000;

  /** Elements of a fact*/
  public enum Column {
    ID, ARG1, ARG2, RELATION
  };

  /** Represents a query template*/
  public static class Template {

    /** Holds the id*/
    protected String id;

    /** Holds the relation as given in the query line*/
    protected String relationString;

    /** Holds the relation (or NULL for variables)*/
    protected Relation relation;

    /** Holds the first and the second fact argument*/
    protected String[] args;

    public Template(String id, String relation, String arg1, String arg2) throws InvalidTripleException {
      super();
      this.id = id;
      relationString = normalizeRelation(relation);
      this.relation = Relations.get(relationString);
      this.args = new String[] { arg1, arg2 };
    }

    public Template(String id, Relation relation, String arg1, String arg2) {
      super();
      this.id = id;
      relationString = relation.name();
      this.relation = relation;
      this.args = new String[] { arg1, arg2 };
    }

    @Override
    public String toString() {
      return id + ":" + args[0] + " " + relationString + " " + args[1];
    }

    /** Returns the relation*/
    public Relation relation() {
      return (relation);
    }

    /** Returns id, arg1, arg2 or relation*/
    public String get(Column c) {
      switch (c) {
        case ID:
          return (id);
        case ARG1:
          return (args[0]);
        case ARG2:
          return (args[1]);
        case RELATION:
          return (relationString);
      }
      return (null);
    }

    /** Tells whether the relation is virtual */
    public boolean isVirtual() {
      return (relation != null && relation instanceof VirtualRelation);
    }

    /** Adds quotes to the n-th argument*/
    public void makeWord(Column arg) {
      int n = arg == Column.ARG1 ? 0 : 1;
      args[n] = '"' + args[n] + '"';
    }

    /** Sets an argument */
    public void setArg(Column arg, String s) {
      int n = arg == Column.ARG1 ? 0 : 1;
      args[n] = s;
    }

    /** Tells whether that argument should be a word */
    public boolean expectWordFor(Column c) {
      if (relation == null) return (false);
      return (c == Column.ARG1 ? relation.domain.isSubClassOf(YagoClasses.STRING) : relation.range.isSubClassOf(YagoClasses.STRING));
    }
  }

  /** Normalizes a relation name*/
  public static String normalizeRelation(String relationName) throws InvalidTripleException {
    if (isVariable(relationName)) return (relationName);
    if (relationName.equalsIgnoreCase("isa")) return (Relations.TYPE.name());
    Relation r = Relations.get(relationName);
    if (r == null) throw new InvalidTripleException("Unknown relation " + relationName);
    return (r.name());
  }

  /** Represents an sql command plus a sequence of variables*/
  public static class SQLCommand {

    /** Holds the command*/
    protected String command;

    /** Holds the variables in the order of the columns of the command*/
    protected List<String> variables;

    public SQLCommand(String command, List<String> variables) {
      super();
      this.command = command;
      this.variables = variables;
    }

    public String getCommand() {
      return command;
    }

    public List<String> getVariables() {
      return variables;
    }

    @Override
    public String toString() {
      return command;
    }
  }

  /** Thrown for problems in a triple*/
  public static class InvalidTripleException extends Exception {

    private static final long serialVersionUID = 4891916879504973582L;

    public InvalidTripleException(String s) {
      super(s);
    }
  }

  /** Tells whether a string starts with "?"*/
  public static boolean isVariable(String s) {
    return (s.startsWith("?"));
  }

  /** Tells whether a string starts with "?@"*/
  public static boolean isSilentVariable(String s) {
    return (s.startsWith("?@"));
  }

  /** Counts fresh variables*/
  public static int varCounter = 0;

  /** returns a new silent variable */
  public static String newSilentVariable(String name) {
    return ("?@" + name + (varCounter++));
  }

  /** returns a variable with the given name*/
  public static String newVariable(String name) {
    return ("?" + name);
  }

  /** Resolves a variable with respect to a mapping or returns the name itself*/
  public static String variableValue(String variableName, Map<String, String> variableValues) {
    String res = variableValues.get(variableName);
    if (res == null) return (variableName);
    else return (res);
  }

  //----------------------------------------------------------------------
  //                Parsing 
  //----------------------------------------------------------------------

  /** Returns e.g. "facts0.arg1" for a variable*/
  public static String sqlIdentifier(String entity, Map<String, String> variables, int table, Column column) {
    if (!isVariable(entity)) return (entity);
    if (variables.containsKey(entity)) return (variables.get(entity));
    String sqlId = column == Column.ID ? ("CAST(f" + table + ".ID AS varchar(255))") : " f" + table + "." + column;
    variables.put(entity, sqlId);
    return (sqlId);
  }

  /** Adds "arg1=arg2" to a collection of conditions*/
  public static void addCondition(Collection<String[]> conditions, String arg1, String arg2) {
    if (arg1.equals(arg2)) return;
    conditions.add(new String[] { arg1, arg2 });
  }

  /** Adds a condition of the form "x=y" for a fact template*/
  public static void addSqlCondition(Template template, int table, Map<String, String> variables, Collection<String[]> conditions) {
    for (Column c : Column.values()) {
      if (c == Column.ID) {
        addCondition(conditions, "CAST(f" + table + ".ID AS varchar(255))", sqlIdentifier(template.get(c), variables, table, c));
      } else {
        addCondition(conditions, " f" + table + "." + c, sqlIdentifier(template.get(c), variables, table, c));
      }
    }
  }

  /** Returns an sqlcommand for a list of templates*/
  public SQLCommand sqlFor(List<Template> templates, Database db) throws InvalidTripleException {

    // Collect all the conditions and all variables
    Map<String, String> variables = new TreeMap<String, String>();
    int numTables = 0;
    Collection<String[]> conditions = new ArrayList<String[]>();
    for (Template template : templates) {
      if (template.isVirtual()) continue;
      addSqlCondition(template, numTables, variables, conditions);
      numTables++;
    }
    if (numTables == 0) throw new InvalidTripleException("No non-virtual relation");

    // Build the SELECT part of the SQL statement, collect all non-silent variables
    StringBuilder result = new StringBuilder("SELECT ");
    List<String> nonSilentVariables = new ArrayList<String>();
    for (String variable : variables.keySet()) {
      if (!isSilentVariable(variable)) {
        nonSilentVariables.add(variable);
        result.append(variables.get(variable)).append(", ");
      }
    }
    if (nonSilentVariables.size() == 0) result.append(1); // SELECT 1 FROM...
    else result.setLength(result.length() - 2);

    // Build the FROM part
    result.append(" FROM ");
    for (int i = 0; i < numTables; i++)
      result.append("facts f").append(i).append(", ");
    if (numTables > 0) result.setLength(result.length() - 2);

    // Build the WHERE part
    result.append(" WHERE\n");
    for (String[] condition : conditions) {
      if (!condition[0].startsWith(" ")) condition[0] = db.format(condition[0]);
      if (!condition[1].startsWith(" ")) condition[1] = db.format(condition[1]);
      result.append(condition[0]).append('=').append(condition[1]).append(" AND\n");
    }
    if (conditions.size() > 0) result.setLength(result.length() - 5);
    return (new SQLCommand(db.limit(result.toString(), MAXSQLRES), nonSilentVariables));
  }

  /** Eats an id from a query line (or returns a fresh variable)*/
  public static String eatId(String template, int[] len) {
    Pattern idPattern = Pattern.compile("\\s*(([\\?\\$][A-Za-z0-9]+)|([0-9]+))\\s*:\\s*");
    Matcher m = idPattern.matcher(template);
    if (!m.find() || m.start() != 0) {
      len[0] = 0;
      return (newSilentVariable("id"));
    }
    len[0] = m.group().length();
    return (m.group(1));
  }

  /** Eats an argument from a query line */
  public static String eatArg(String template, int[] len, boolean tabSeparated) {
    if (template.length() == 0) {
      len[0] = 0;
      return (null);
    }
    if (template.startsWith("\"")) {
      int end = template.indexOf('"', 2);
      if (end == -1) return (null);
      len[0] = end + 1;
      return (template.substring(1, end).trim());
    }
    int end = template.indexOf(tabSeparated ? '\t' : ' ');
    if (end == -1) end = template.length();
    len[0] = end;
    return (template.substring(0, end).trim());
  }

  /** Eats a relation from a query line */
  public static String eatRelation(String template, int[] len, boolean tabSeparated) {
    if (template.length() == 0) {
      len[0] = 0;
      return (null);
    }
    int end = template.indexOf(tabSeparated ? '\t' : ' ');
    if (end == -1) end = template.length();
    len[0] = end;
    String ret = template.substring(0, end).trim();
    if (ret.equalsIgnoreCase("is")) {
      ret = "isa";
      end = template.indexOf(' ', end + 1);
      len[0] = end;
    }
    if (ret.equalsIgnoreCase("isa")) ret = Relations.TYPE.name();
    return (ret);
  }

  /** Adds templates to a list for a query line*/
  public static void parseLine(List<Template> result, String line) throws InvalidTripleException {
    int[] len = new int[1];
    boolean tabSeparated = line.contains("\t");

    // Get first triple
    String id = eatId(line, len);
    line = line.substring(len[0]).trim();
    String arg1 = eatArg(line, len, tabSeparated);
    if (arg1 == null) return;
    line = line.substring(len[0]).trim();
    String relation = eatRelation(line, len, tabSeparated);
    line = line.substring(len[0]).trim();
    if (relation == null) {
      result.add(new Template(id, newVariable("relation"), arg1, newVariable("x")));
      return;
    }
    String arg2 = eatArg(line, len, tabSeparated);
    line = line.substring(len[0]).trim();
    if (arg2 == null) {
      result.add(new Template(id, relation, arg1, newVariable("x")));
      return;
    }
    result.add(new Template(id, relation, arg1, arg2));

    // Get the other triples
    while (true) {
      relation = eatRelation(line, len, tabSeparated);
      line = line.substring(len[0]).trim();
      if (relation == null) return;
      arg2 = eatArg(line, len, tabSeparated);
      line = line.substring(len[0]).trim();
      if (arg2 == null) throw new InvalidTripleException(line);
      String newId = newSilentVariable("id");
      result.add(new Template(newId, relation, id, arg2));
      id = newId;
    }
  }

  /** Resolves MEANS and adds quotation marks where necessary*/
  public static void resolveMeans(List<Template> templates) {
    // We cannot use an iterator here, because we modify the list
    int numTemplates = templates.size();
    for (int i = 0; i < numTemplates; i++) {
      Template template = templates.get(i);
      for (Column c : new Column[] { Column.ARG1, Column.ARG2 }) {
        if (template.get(c).matches("\\d.*") || isVariable(template.get(c)) || template.get(c).contains("_")) {
          continue;
        }
        if (template.expectWordFor(c)) {
          template.makeWord(c);
          continue;
        }
        if (template.get(c).startsWith("'")) {
          template.setArg(c, template.get(c).replace("'", ""));
          continue;
        }
        String variable = newVariable(template.get(c));
        Template means = new Template(newSilentVariable("id"), Relations.MEANS, template.get(c), variable);
        means.makeWord(Column.ARG1);
        template.setArg(c, variable);
        templates.add(means);
      }
    }
  }

  /** Returns templates for a query*/
  public static List<Template> templatesFor(String input) throws InvalidTripleException {
    List<Template> list = new ArrayList<Template>();
    for (String line : input.split("\n|;")) {
      parseLine(list, line);
    }
    resolveMeans(list);
    return (list);
  }

  /** returns templates for a list of 3-element query lines with implict ids ?id0, ?id1, ?id2,...*/
  public static List<Template> templatesFor(String[][] queryLines) throws InvalidTripleException {
    List<Template> list = new ArrayList<Template>();
    int id = 0;
    for (String[] line : queryLines) {
      if (line == null || line.length != 3 || line[0] == null || line[1] == null || line[2] == null || line[0].length() == 0) continue;
      list.add(new Template("?id" + id, line[1], line[0], line[2]));
      id++;
    }
    resolveMeans(list);
    return (list);
  }

  // ------------------------------------------------------------------
  //                  Query solving
  // ------------------------------------------------------------------

  /** Checks the virtual relations in an answer */
  public static boolean checkVirtual(List<Template> templates, Map<String, String> variables) throws InvalidTripleException {
    for (Template template : templates) {
      if (template.isVirtual()) {
        VirtualRelation r = (VirtualRelation) template.relation();
        String arg1 = variableValue(template.get(Column.ARG1), variables);
        String arg2 = variableValue(template.get(Column.ARG2), variables);
        if (isVariable(arg2) || isVariable(arg1)) throw new InvalidTripleException("Unbound argument for virtual relation " + r);
        if (!r.holds(arg1, arg2)) return (false);
      }
    }
    return (true);
  }

  /** Iterates over variable bindings for a list of templates*/
  public PeekIterator<Map<String, String>> solutions(final List<Template> templates, final Database db) throws InvalidTripleException {

    final SQLCommand sql = sqlFor(templates, db);
    return (new PeekIterator<Map<String, String>>() {

      ResultSet rs = null;

      public Map<String, String> internalNext() throws Exception {
        if (rs == null) rs = db.query(sql.getCommand());
        Map<String, String> variables;
        do {
          if (!rs.next()) return (null);
          variables = new TreeMap<String, String>();
          int i = 1;
          for (String var : sql.getVariables()) {
            variables.put(var, rs.getString(i++));
          }
        } while (!checkVirtual(templates, variables));
        return variables;
      }

      @Override
      public void close() {
        Database.close(rs);
      }
    });
  }

  /** Returns up to n solutions for a query */
  public Set<Map<String, String>> solutions(List<Template> templates, final Database db, int n) throws InvalidTripleException {
    Set<Map<String, String>> solutions = new HashSet<Map<String, String>>();
    Iterator<Map<String, String>> it = solutions(templates, db);
    while (it.hasNext() && solutions.size() < n) {
      solutions.add(it.next());
    }
    return (solutions);
  }

  /** Returns the solutions for a query */
  public Iterator<Map<String, String>> solutions(String query, final Database db) throws InvalidTripleException {
    return (solutions(templatesFor(query), db));
  }

  /** Returns up to n solutions for a query */
  public Set<Map<String, String>> solutions(String query, final Database db, int n) throws InvalidTripleException {
    return (solutions(templatesFor(query), db, n));
  }

  /** Returns the solutions for a query */
  public Iterator<Map<String, String>> solutions(String[][] queryLines, final Database db) throws InvalidTripleException {
    return (solutions(templatesFor(queryLines), db));
  }

  /** Returns up to n solutions for a query */
  public Set<Map<String, String>> solutions(String[][] queryLines, final Database db, int n) throws InvalidTripleException {
    return (solutions(templatesFor(queryLines), db, n));
  }

  // -----------------------------------------------------------
  //                     Test
  // -----------------------------------------------------------

  /** request information instead of using yago.ini file*/
  public static void requestInformation() throws Exception {
    boolean ok = false;
    while (!ok) {
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
    }
    while (db == null) {
      D.println("");
      D.println("Which Database are you using (quit to end)");
      D.println("1. Oracle");
      D.println("2. MySql");
      D.println("3. Postgres");
      D.println("4. quit");
      int choice = 0;
      while (choice <= 0 || choice >= 5) {
        D.pl("input:> ");
        try {
          String input = D.r().trim();
          if (input.equals("quit")) {
            D.println("program ended");
            return;
          }
          choice = Integer.parseInt(input);
          if (choice == 4) {
            D.println("program ended");
            return;
          }
        } catch (Exception e) {
          choice = 0;
        }
      }
      String system = null;
      if (choice == 1) {
        system = "ORACLE";
        if (!OracleDatabase.checkJar()) {
          D.pl("**********************************************************************\n");
          D.pl("You are missing the oracle jdbc, see readme_converters.txt\n");
          D.pl("**********************************************************************\n");
          return;
        }
      } else if (choice == 2) {
        system = "MYSQL";
        if (!MySQLDatabase.checkJar()) {
          D.pl("**********************************************************************\n");
          D.pl("You are missing the mysql jdbc, see readme_converters.txt\n");
          D.pl("**********************************************************************\n");
          return;
        }
      } else if (choice == 3) {
        system = "POSTGRES";
        if (!PostgresDatabase.checkJar()) {
          D.pl("*************************************************************************\n");
          D.pl("You are missing the postgres jdbc, see readme_converters.txtv");
          D.pl("*************************************************************************\n");
          return;
        }
      }
      D.pl("Please enter user:> ");
      String user = D.r().trim();
      D.pl("Please enter password:> ");
      String password = D.r().trim();
      D.pl("Please enter host:> ");
      String host = D.r().trim();
      D.pl("Please enter port:> ");
      String port = D.r().trim();
      String schema = null;
      String inst = null;
      if (choice == 1) {
        D.pl("Please enter oracle SID:> ");
        inst = D.r().trim();
      }
      if (choice == 2) {
        D.pl("Please enter mysql database:> ");
        schema = D.r().trim();
      }
      if (choice == 3) {
        D.pl("Please enter postgres schema:> ");
        schema = D.r().trim();
      }
      D.println("Trying to Connect to the database");
      try {
        db = DBConfig.getDatabase(system, user, password, host, port, schema, inst);
        D.println("Connected to", db);
      } catch (Exception e) {
        D.println("Failed to Connect to database");
        db = null;
      }
    }
    query();
  }

  private static Database db = null;

  public static void main(String[] args) throws Exception {
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFile(f);
    } else {
      requestInformation();
    }
  }

  /** reads the yago.ini file*/
  public static void fromIniFile(File initFile) throws Exception {
    Parameters.init(initFile);
    Announce.doing("Connecting to database");
    db = DBConfig.getDatabase(initFile.getName());
    DataGraphManager.init();
    Announce.message("Connected to", db);
    query();
  }

  /** Tests the Reification Query processor*/
  public static void query() throws Exception {
    QueryProcessor qp = new QueryProcessor();
    D.p("Example query: \"Angela Merkel\" means ?x; ?x type ?y");
    while (true) {
      D.p("Enter query lines, followed by a blank line. Type blank line to quit.");
      String line = "";
      while (true) {
        String input = D.r().trim();
        if (input == null || input.length() == 0) break;
        line += "; " + input;
      }
      if (line.length() == 0) break;
      try {
        List<Template> list = templatesFor(line);
        D.p(list);
        int count = 0;
        for (Map<String, String> solution : qp.solutions(list, db, 100)) {
          count++;
          D.p(solution);
        }
        if (count == 0) {
          D.println("No Results found");
        }
        D.println("");
      } catch (Exception e) {
        D.p(e.getMessage());
      }
    }
    D.p("Closing database");
    db.close();
  }
}
