package converters;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.Map;
import java.util.TreeMap;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.datatypes.FinalMap;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.FileSet;
import javatools.filehandlers.UTF8Reader;
import javatools.parsers.Char;
import basics.datagraph.DataGraphManager;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.YagoClass;
import basics.datagraph.YagoClasses;

/**
 * This class is part of the YAGO converters (http://mpii.de/yago). 
 * It is licensed under the Creative Commons Attribution License 
 * (http://creativecommons.org/licenses/by/3.0) 
 * by the YAGO team (http://mpii.de/yago). 
 * 
 * This class produces the RDFS version of YAGO
 * 
 * @author Fabian M. Suchanek*/
public class RDFSConverter {

  /** Holds the YAGO namespace (need not be a URL or valid URL)*/
  public static final String ns = "http://www.mpii.de/yago/resource";

  /** holds the YAGO folder*/
  public static File yagoFolder;

  /** holds the output folder*/
  public static File outputFolder;

  /** Tells whether the deductive closure shall be converted as well*/
  public static boolean deductiveClosure;

  /** Tells whether facts about facts shall be included*/
  public static boolean includeFactFacts;

  /** All rdfs datatypes*/
  public static enum RDFSDatatype {
    rboolean, rdouble, rdecimal, rnonNegativeInteger, rstring, rdate, rgYear, rduration, rRESOURCE;

    @Override
    public String toString() {
      return "\"&d;" + name().substring(1) + '"';
    }

    public String asLiteral(String literal) {
      if (literal.startsWith("\"")) literal = Char.cutLast(literal.substring(1));
      return (Char.encodeAmpersand(literal));
    }
  };

  /** Maps Yago classes to RDFS types */
  private static Map<YagoClass, RDFSDatatype> yago2rdfs = null;

  /**
   * prepares the specialRelationNames and yago2rdfs mapping. must be run after the Datagraph has been initialised.
   */
  private static void prepare() {
    specialRelationNames = new FinalMap<Relation, String>(Relations.MEANS, "rdfs:label", Relations.TYPE, "rdf:type", Relations.SUBCLASSOF, "rdfs:subClassOf", Relations.SUBPROPERTYOF, "rdfs:subPropertyOf", Relations.DOMAIN, "rdfs:domain",
        Relations.RANGE, "rdfs:range", Relations.FAMILYNAME, "y:hasFamilyName", Relations.GIVENNAME, "y:hasGivenName", Relations.NATIVENAME, "y:hasNativeName");
    yago2rdfs = new FinalMap<YagoClass, RDFSDatatype>(new Object[] {
        YagoClasses.BOOLEAN,
        RDFSDatatype.rboolean,
        YagoClasses.NUMBER,
        RDFSDatatype.rdouble,
        YagoClasses.RATIONAL,
        RDFSDatatype.rdouble,
        YagoClasses.INTEGER,
        RDFSDatatype.rdecimal,
        YagoClasses.NONNEGATIVEINTEGER,
        RDFSDatatype.rnonNegativeInteger,
        YagoClasses.STRING,
        RDFSDatatype.rstring,
        YagoClasses.WORD,
        RDFSDatatype.rstring,
        YagoClasses.TLD,
        RDFSDatatype.rstring,
        YagoClasses.CHAR,
        RDFSDatatype.rstring,
        YagoClasses.TIMEINTERVAL,
        RDFSDatatype.rdate,
        YagoClasses.TIMEPOINT,
        RDFSDatatype.rdate,
        YagoClasses.DATE,
        RDFSDatatype.rdate,
        YagoClasses.YEAR,
        RDFSDatatype.rgYear,
        YagoClasses.DURATION,
        RDFSDatatype.rduration,
        YagoClasses.IDENTIFIER,
        RDFSDatatype.rstring,
        YagoClasses.CALLINGCODE,
        RDFSDatatype.rstring,
        YagoClasses.ISBN,
        RDFSDatatype.rstring,
        YagoClasses.URL });
  }

  /** Returns the RDFSType for a YagoClass (or RESOURCE if it's a resource)*/
  public static RDFSDatatype rdfsTypeforYagoClass(YagoClass y) {
    if (yago2rdfs.containsKey(y)) return (yago2rdfs.get(y));
    else return (RDFSDatatype.rRESOURCE);
  }

  /** Formats a Yago entity as a URI*/
  public static String asURI(String entity) {
    entity = Char.encodeAmpersand(entity);
    if (!entity.startsWith("http://")) entity = "&y;" + entity;
    return ('"' + entity + '"');
  }

  /** Maps Yago relation names to their RDFS equivalent*/
  public static Map<Relation, String> specialRelationNames = null;

  /** Creates a writer for a relation and an extractor filename, writes header */
  public static Writer writerForRDFS(Relation yagoRelation, File f) throws IOException {
    String rdfsFileName = yagoRelation.name() + '_' + FileSet.newExtension(f.getName(), "rdfs");
    Writer out = new BufferedWriter(new FileWriter(new File(outputFolder, rdfsFileName)));
    out.write("<?xml version=\"1.0\"?>\n");
    out.write("<!DOCTYPE rdf:RDF [<!ENTITY d \"http://www.w3.org/2001/XMLSchema#\">\n");
    out.write("                   <!ENTITY y \"" + ns + "/\">]>\n");
    out.write("<rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"\n");
    out.write("         xml:base=\"" + ns + "\"\n");
    out.write("         xmlns:rdfs=\"http://www.w3.org/2000/01/rdf-schema#\"\n");
    out.write("         xmlns:y=\"" + ns + "/\">\n");
    return (out);
  }

  /** Creates a writer for a relation and an extractor filename, writes header */
  public static Writer writerForN3(Relation yagoRelation, File f) throws IOException {
    String rdfsFileName = yagoRelation == null ? "yago.n3" : yagoRelation.name() + '_' + FileSet.newExtension(f.getName(), "n3");
    Writer out = new BufferedWriter(new FileWriter(new File(outputFolder, rdfsFileName)));
    D.writeln(out, "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .");
    D.writeln(out, "@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .");
    D.writeln(out, "@prefix y: <http://mpii.de/yago/resource/> .");
    D.writeln(out, "@prefix x: <http://www.w3.org/2001/XMLSchema#> .");
    D.writeln(out, "@base <http://mpii.de/yago/resource/> .");
    return (out);
  }

  /** Language codes */
  public static Map<String, String> languageCodes = new FinalMap<String, String>("Afar_language", "aa", "Abkhazian_language", "ab", "Avestan_language", "ae", "Afrikaans_language", "af", "Akan_language", "ak", "Amharic_language", "am",
      "Aragonese_language", "an", "Arabic_language", "ar", "Assamese_language", "as", "Avaric_language", "av", "Aymara_language", "ay", "Azerbaijani_language", "az", "Bashkir_language", "ba", "Belarusian_language", "be", "Bulgarian_language", "bg",
      "Bihari_language", "bh", "Bislama_language", "bi", "Bambara_language", "bm", "Bengali_language", "bn", "Tibetan_language", "bo", "Breton_language", "br", "Bosnian_language", "bs", "Catalan_language", "ca", "Chechen_language", "ce",
      "Chamorro_language", "ch", "Corsican_language", "co", "Cree_language", "cr", "Czech_language", "cs", "Church_language", "cu", "Chuvash_language", "cv", "Welsh_language", "cy", "Danish_language", "da", "German_language", "de",
      "Divehi_language", "dv", "Dzongkha_language", "dz", "Ewe_language", "ee", "Greek_language", "el", "English_language", "en", "Esperanto_language", "eo", "Spanish_language", "es", "Estonian_language", "et", "Basque_language", "eu",
      "Persian_language", "fa", "Fulah_language", "ff", "Finnish_language", "fi", "Fijian_language", "fj", "Faroese_language", "fo", "French_language", "fr", "Western Frisian_language", "fy", "Irish_language", "ga", "Scottish_language", "gd",
      "Galician_language", "gl", "Guaran�_language", "gn", "Gujarati_language", "gu", "Manx_language", "gv", "Hausa_language", "ha", "Hebrew_language", "he", "Hindi_language", "hi", "Hiri_language", "ho", "Croatian_language", "hr",
      "Haitian_language", "ht", "Hungarian_language", "hu", "Armenian_language", "hy", "Herero_language", "hz", "Interlingua_language", "ia", "Indonesian_language", "id", "Interlingue_language", "ie", "Igbo_language", "ig", "Sichuan_language", "ii",
      "Inupiaq_language", "ik", "Ido_language", "io", "Icelandic_language", "is", "Italian_language", "it", "Inuktitut_language", "iu", "Japanese_language", "ja", "Javanese_language", "jv", "Georgian_language", "ka", "Kongo_language", "kg",
      "Kikuyu_language", "ki", "Kwanyama_language", "kj", "Kazakh_language", "kk", "Kalaallisut_language", "kl", "Khmer_language", "km", "Kannada_language", "kn", "Korean_language", "ko", "Kanuri_language", "kr", "Kashmiri_language", "ks",
      "Kurdish_language", "ku", "Komi_language", "kv", "Cornish_language", "kw", "Kirghiz_language", "ky", "Latin_language", "la", "Luxembourgish_language", "lb", "Ganda_language", "lg", "Limburgish_language", "li", "Lingala_language", "ln",
      "Lao_language", "lo", "Lithuanian_language", "lt", "Luba-Katanga_language", "lu", "Latvian_language", "lv", "Malagasy_language", "mg", "Marshallese_language", "mh", "Maori_language", "mi", "Macedonian_language", "mk", "Malayalam_language",
      "ml", "Mongolian_language", "mn", "Moldavian_language", "mo", "Marathi_language", "mr", "Malay_language", "ms", "Maltese_language", "mt", "Burmese_language", "my", "Nauru_language", "na", "North_language", "nd", "Nepali_language", "ne",
      "Ndonga_language", "ng", "Dutch_language", "nl", "Norwegian_language", "no", "South_language", "nr", "Navajo_language", "nv", "Chichewa_language", "ny", "Occitan_language", "oc", "Ojibwa_language", "oj", "Oromo_language", "om",
      "Oriya_language", "or", "Ossetian_language", "os", "Panjabi_language", "pa", "Pali_language", "pi", "Polish_language", "pl", "Pashto_language", "ps", "Portuguese_language", "pt", "Quechua_language", "qu", "Raeto-Romance_language", "rm",
      "Kirundi_language", "rn", "Romanian_language", "ro", "Russian_language", "ru", "Kinyarwanda_language", "rw", "Rusyn_language", "ry", "Sanskrit_language", "sa", "Sardinian_language", "sc", "Sindhi_language", "sd", "Northern_language", "se",
      "Sango_language", "sg", "Serbo-Croatian_language", "sh", "Sinhalese_language", "si", "Slovak_language", "sk", "Slovenian_language", "sl", "Samoan_language", "sm", "Shona_language", "sn", "Somali_language", "so", "Albanian_language", "sq",
      "Serbian_language", "sr", "Swati_language", "ss", "Sotho_language", "st", "Sundanese_language", "su", "Swedish_language", "sv", "Swahili_language", "sw", "Tamil_language", "ta", "Telugu_language", "te", "Tajik_language", "tg", "Thai_language",
      "th", "Tigrinya_language", "ti", "Turkmen_language", "tk", "Tagalog_language", "tl", "Tswana_language", "tn", "Tonga_language", "to", "Turkish_language", "tr", "Tsonga_language", "ts", "Tatar_language", "tt", "Twi_language", "tw",
      "Tahitian_language", "ty", "Uighur_language", "ug", "Ukrainian_language", "uk", "Urdu_language", "ur", "Uzbek_language", "uz", "Venda_language", "ve", "Vietnamese_language", "vi", "Volap�k_language", "vo", "Walloon_language", "wa",
      "Wolof_language", "wo", "Xhosa_language", "xh", "Yiddish_language", "yi", "Yoruba_language", "yo", "Zhuang_language", "za", "Chinese_language", "zh", "Zulu_language", "zu");

  /** Converts all facts to RDFS */
  public static void convertFactstoRDFS(boolean test) throws IOException {
    Announce.doing("Converting facts to RDFS");
    // Handle all relations except for ISCALLED and INLANGUAGE
    for (File yagoFactFile : new File(yagoFolder, "facts").listFiles()) {
      if (!yagoFactFile.isDirectory()) continue;
      Relation yagoRelation = Relations.get(yagoFactFile.getName());
      if (yagoRelation == null) {
        Announce.message("No relation found for folder", yagoFactFile.getName());
        continue;
      }
      if (yagoRelation == Relations.ISCALLED || yagoRelation == Relations.INLANGUAGE) continue;
      if (!includeFactFacts && yagoRelation.domain == YagoClasses.FACT) continue;
      String yagoRelationName = yagoRelation.name();
      Announce.doing("Converting", yagoRelation);
      RDFSDatatype objectType = rdfsTypeforYagoClass(yagoRelation.range);
      boolean inverted = false;
      if (rdfsTypeforYagoClass(yagoRelation.domain) != RDFSDatatype.rRESOURCE) {
        inverted = true;
        objectType = rdfsTypeforYagoClass(yagoRelation.domain);
      }
      if (specialRelationNames.containsKey(yagoRelation)) yagoRelationName = specialRelationNames.get(yagoRelation);
      else yagoRelationName = "y:" + yagoRelationName;
      boolean arg1isClass = inverted ? yagoRelation.range.isSubClassOf(YagoClasses.CLASS) : yagoRelation.domain.isSubClassOf(YagoClasses.CLASS);

      for (File f : yagoFactFile.listFiles()) {
        if (!deductiveClosure && f.getName().equalsIgnoreCase("IsAExtractor.txt")) continue;
        Writer out = writerForRDFS(yagoRelation, f);
        int counter = 0;
        FileLines linesOfF = new FileLines(new UTF8Reader(f, "Parsing " + f.getName()));
        for (String l : linesOfF) {
          if (test && counter++ > 100) break;
          String[] split = l.split("\t");
          if (split.length != 4) continue;
          String id = split[0];
          String arg1 = split[inverted ? 2 : 1];
          String arg2 = split[inverted ? 1 : 2];
          String confidence = split[3];
          out.write(arg1isClass ? "<rdfs:Class" : "<rdf:Description");
          if (yagoRelation.domain == YagoClasses.FACT && !inverted) out.write(" rdf:about=\"#f" + arg1 + "\"");
          else out.write(" rdf:about=" + asURI(arg1));
          out.write(">");
          out.write("<" + yagoRelationName + " rdf:ID=\"f" + id + "\"");
          if (objectType == RDFSDatatype.rRESOURCE) {
            out.write(" rdf:resource=" + asURI(arg2) + "/>");
          } else {
            out.write(" rdf:datatype=" + objectType + ">");
            out.write(objectType.asLiteral(arg2));
            out.write("</" + yagoRelationName + ">");
          }
          out.write(arg1isClass ? "</rdfs:Class>\n" : "</rdf:Description>\n");
          out.write("  <rdf:Description rdf:about=\"#f" + id + "\">");
          out.write("<y:confidence rdf:datatype=" + RDFSDatatype.rdouble + ">" + confidence + "</y:confidence>");
          out.write("</rdf:Description>\n");
        }
        linesOfF.close();
        out.write("</rdf:RDF>\n");
        out.close();
      }
      Announce.done();
    }

    // Handle iscalled and inlanguage
    Writer out = writerForRDFS(Relations.ISCALLED, new File("ArticleExtractor.txt"));
    FileLines iscalled = new FileLines(new UTF8Reader(FileSet.file(yagoFolder, "facts", Relations.ISCALLED.toString(), "ArticleExtractor.txt"), "Parsing isCalled"));
    FileLines inlanguage = new FileLines(new UTF8Reader(FileSet.file(yagoFolder, "facts", Relations.INLANGUAGE.toString(), "ArticleExtractor.txt")));
    int counter = 0;
    while (iscalled.hasNext()) {
      if (test && counter++ > 100) break;
      Map<String, String> id2name = new TreeMap<String, String>();
      String[] iscalledSplit = iscalled.peek().split("\t");
      String currentEntity = iscalledSplit[1];
      while (currentEntity.equals(iscalledSplit[1])) {
        id2name.put(iscalledSplit[0], iscalledSplit[2]);
        iscalled.next();
        if (!iscalled.hasNext()) break;
        iscalledSplit = iscalled.peek().split("\t");
      }
      out.write("<rdf:Description rdf:about=" + asURI(currentEntity) + ">\n");
      while (inlanguage.hasNext()) {
        String[] inlanguageSplit = inlanguage.peek().split("\t");
        if (!id2name.containsKey(inlanguageSplit[1])) break;
        out.write("  <rdfs:label rdf:ID=\"f" + inlanguageSplit[0] + "\" xml:lang=\"" + languageCodes.get(inlanguageSplit[2]) + "\">" + RDFSDatatype.rstring.asLiteral(id2name.get(inlanguageSplit[1])) + "</rdfs:label>\n");
        inlanguage.next();
      }
      out.write("</rdf:Description>\n");
    }
    iscalled.close();
    inlanguage.close();
    out.write("</rdf:RDF>\n");
    out.close();
    Announce.done();
  }

  /** Creates a N3 identifier for a YAGO identifier.*/
  public static String n3NameFor(String yagoName) {
    if (yagoName.startsWith("\"http:")) return ("<" + Char.cutLast(yagoName.substring(1)) + ">");
    if (yagoName.startsWith("http:")) return ("<" + yagoName + ">");
    return ("<" + Char.encodeURIPathComponent(yagoName) + ">");
  }

  /** Creates a N3 string */
  public static String n3String(String string) {
    if (string.startsWith("\"")) string = Char.cutLast(string.substring(1));
    return ('"' + Char.encodeBackslash(string) + '"');
  }

  /** Converts all facts to N3 */
  public static void convertFactstoN3(boolean test, boolean oneFile) throws IOException {
    Announce.doing("Converting facts to N3" + (test ? " in test mode" : ""));
    Writer out = null;
    if (oneFile) out = writerForN3(null, null);
    // Handle all relations except for ISCALLED and INLANGUAGE
    for (File yagoFactFile : new File(yagoFolder, "facts").listFiles()) {
      if (!yagoFactFile.isDirectory()) continue;
      Relation yagoRelation = Relations.get(yagoFactFile.getName());
      if (yagoRelation == null) {
        Announce.message("No relation found for folder", yagoFactFile.getName());
        continue;
      }
      if (yagoRelation == Relations.ISCALLED || yagoRelation == Relations.INLANGUAGE) continue;
      if (!includeFactFacts && yagoRelation.domain == YagoClasses.FACT) continue;
      String yagoRelationName = yagoRelation.name();
      Announce.doing("Converting", yagoRelation);
      RDFSDatatype objectType = rdfsTypeforYagoClass(yagoRelation.range);
      boolean inverted = false;
      if (rdfsTypeforYagoClass(yagoRelation.domain) != RDFSDatatype.rRESOURCE) {
        inverted = true;
        objectType = rdfsTypeforYagoClass(yagoRelation.domain);
      }
      if (specialRelationNames.containsKey(yagoRelation)) yagoRelationName = specialRelationNames.get(yagoRelation);
      else yagoRelationName = "y:" + yagoRelationName;

      for (File f : yagoFactFile.listFiles()) {
        if (!deductiveClosure && f.getName().equalsIgnoreCase("IsAExtractor.txt")) continue;
        if (!oneFile) out = writerForN3(yagoRelation, f);
        int counter = 0;
        FileLines linesOfF = new FileLines(new UTF8Reader(f, "Parsing " + f.getName()));
        for (String l : linesOfF) {
          if (test && counter++ > 100) break;
          String[] split = l.split("\t");
          if (split.length != 4) continue;
          String arg1 = split[inverted ? 2 : 1];
          String arg2 = split[inverted ? 1 : 2];
          out.write(n3NameFor(arg1) + " " + yagoRelationName + " ");
          switch (objectType) {
            case rboolean:
            case rdecimal:
            case rgYear:
            case rnonNegativeInteger:
            case rdouble:
              out.write('"' + arg2 + '"');
              break;
            case rdate:
            case rduration:
            case rRESOURCE:
              out.write(n3NameFor(arg2));
              break;
            case rstring:
              out.write(n3String(arg2));
              break;
          }
          out.write(" .\n");
        }
        linesOfF.close();
        if (!oneFile) out.close();
      }
      Announce.done();
    }

    // Handle iscalled and inlanguage
    if (!oneFile) out = writerForN3(Relations.ISCALLED, new File("ArticleExtractor.txt"));
    FileLines iscalled = new FileLines(new UTF8Reader(FileSet.file(yagoFolder, "facts", Relations.ISCALLED.toString(), "ArticleExtractor.txt"), "Parsing isCalled"));
    FileLines inlanguage = new FileLines(new UTF8Reader(FileSet.file(yagoFolder, "facts", Relations.INLANGUAGE.toString(), "ArticleExtractor.txt")));
    int counter = 0;
    while (iscalled.hasNext()) {
      if (test && counter++ > 100) break;
      Map<String, String> id2name = new TreeMap<String, String>();
      String[] iscalledSplit = iscalled.peek().split("\t");
      String currentEntity = iscalledSplit[1];
      while (currentEntity.equals(iscalledSplit[1])) {
        id2name.put(iscalledSplit[0], iscalledSplit[2]);
        iscalled.next();
        if (!iscalled.hasNext()) break;
        iscalledSplit = iscalled.peek().split("\t");
      }

      while (inlanguage.hasNext()) {
        String[] inlanguageSplit = inlanguage.peek().split("\t");
        if (!id2name.containsKey(inlanguageSplit[1])) break;
        out.write(n3NameFor(currentEntity) + " rdfs:label ");
        out.write(n3String(id2name.get(inlanguageSplit[1])) + "@" + languageCodes.get(inlanguageSplit[2]));
        out.write(" .\n");
        inlanguage.next();
      }
    }
    iscalled.close();
    //inlanguage.close();
    out.close();
    Announce.done();
  }

  /** request information instead of using yago.ini file*/
  public static void requestInformation() throws Exception {
    while (yagoFolder == null) {
      D.println("Please enter a yago Folder directory containing the yago facts and entities (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      yagoFolder = new File(path);
      if (!yagoFolder.exists()) {
        D.println("yago Folder does not exist, please try again (quit to end)");
        yagoFolder = null;
      } else {
        if (!new File(yagoFolder, "facts").exists()) {
          D.println(path + " does not have a sub-folder 'facts', please try again (quit to end)");
          yagoFolder = null;
        }
        if (!new File(yagoFolder, "entities").exists()) {
          D.println(path + " does not have a sub-folder 'entities', please try again (quit to end)");
          yagoFolder = null;
        }
      }
    }
    while (outputFolder == null) {
      D.println("Please enter a target directory (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      outputFolder = new File(path);
      if (!outputFolder.exists()) {
        if (!outputFolder.mkdir()) {
          outputFolder = null;
        }
      }
    }
    boolean ok = false;
    while (!ok) {
      D.println("");
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
      if (ok) {
        prepare();
      }
    }

    String ans = null;
    while (ans == null) {
      D.println("");
      D.println("Would you like to add deductive closer of the form");
      D.println("(x,type,c1), (c1, subclassof, c2) => (x, type, c2)");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        deductiveClosure = true;
      } else if (ans.equalsIgnoreCase("n")) {
        deductiveClosure = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    ans = null;
    while (ans == null) {
      D.println("Do you want to include facts about facts");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        includeFactFacts = true;
      } else if (ans.equalsIgnoreCase("n")) {
        includeFactFacts = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    convertFactstoRDFS(false);
  }

  /** reads the yago.ini file*/
  public static void fromIniFile(File initFile) throws Exception {
    Parameters.init(initFile);
    DataGraphManager.init();
    prepare();
    Parameters.ensureParameters("yagoFolder - the YAGO data folder", "rdfsFolder - the folder where the RDFS data shall go");
    yagoFolder = Parameters.getFile("yagoFolder");
    outputFolder = Parameters.getFile("rdfsFolder");
    deductiveClosure = Parameters.isDefined("deductiveClosure") && Parameters.getBoolean("deductiveClosure");
    includeFactFacts = Parameters.isDefined("includeFactFacts") && Parameters.getBoolean("includeFactFacts");
    Announce.doing("Creating rdfs folder");
    outputFolder.mkdir();
    Announce.done();
    Announce.message("Deductive closure:", deductiveClosure);
    Announce.message("Include fact facts:", includeFactFacts);
    convertFactstoRDFS(false);
  }

  /** Produces the RDFS version of Yago*/
  public static void main(String[] args) throws Exception {
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFile(f);
    } else {
      requestInformation();
    }
  }

  /** reads the yago.ini file*/
  public static void fromIniFileN3(File initFile) throws Exception {
    Parameters.init(initFile);
    DataGraphManager.init();
    prepare();
    Parameters.ensureParameters("yagoFolder - the YAGO data folder", "n3Folder - the folder where the N3 data shall go", "N3oneFile - target directory should contain only one file");
    yagoFolder = Parameters.getFile("yagoFolder");
    outputFolder = Parameters.getFile("n3Folder");
    boolean oneFile = Boolean.parseBoolean(Parameters.get("N3oneFile"));
    deductiveClosure = Parameters.isDefined("deductiveClosure") && Parameters.getBoolean("deductiveClosure");
    includeFactFacts = false;
    Announce.doing("Creating n3 folder");
    outputFolder.mkdir();
    Announce.done();
    Announce.message("Deductive closure:", deductiveClosure);
    Announce.message("Include fact facts:", includeFactFacts);
    convertFactstoN3(false, oneFile);
  }

  /** Produces N3 output for YAGO*/
  public static void mainN3(String[] args) throws Exception {
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFileN3(f);
    } else {
      requestInformationN3();
    }
  }

  /** request information instead of using yago.ini file*/
  public static void requestInformationN3() throws Exception {
    while (yagoFolder == null) {
      D.println("Please enter a yago Folder directory containing the yago facts and entities (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      yagoFolder = new File(path);
      if (!yagoFolder.exists()) {
        D.println("yago Folder does not exist, please try again (quit to end)");
        yagoFolder = null;
      } else {
        if (!new File(yagoFolder, "facts").exists()) {
          D.println(path + " does not have a sub-folder 'facts', please try again (quit to end)");
          yagoFolder = null;
        }
        if (!new File(yagoFolder, "entities").exists()) {
          D.println(path + " does not have a sub-folder 'entities', please try again (quit to end)");
          yagoFolder = null;
        }
      }
    }
    while (outputFolder == null) {
      D.println("Please enter a target directory (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      outputFolder = new File(path);
      if (!outputFolder.exists()) {
        if (!outputFolder.mkdir()) {
          outputFolder = null;
        }
      }
    }
    boolean ok = false;
    while (!ok) {
      D.println("");
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
      if (ok) {
        prepare();
      }
    }

    String ans = null;
    while (ans == null) {
      D.println("");
      D.println("Would you like to add deductive closer of the form");
      D.println("(x,type,c1), (c1, subclassof, c2) => (x, type, c2)");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        deductiveClosure = true;
      } else if (ans.equalsIgnoreCase("n")) {
        deductiveClosure = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    ans = null;
    boolean oneFile = false;
    while (ans == null) {
      D.println("");
      D.println("Would you the output to be one file");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        oneFile = true;
      } else if (ans.equalsIgnoreCase("n")) {
        oneFile = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    includeFactFacts = false;
    convertFactstoN3(false, oneFile);
  }

}
