package converters;

import java.io.File;
import java.sql.Types;
import java.util.Arrays;
import java.util.List;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.database.Database;
import javatools.database.MySQLDatabase;
import javatools.database.OracleDatabase;
import javatools.database.PostgresDatabase;
import javatools.database.Database.Inserter;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.UTF8Reader;
import basics.configuration.DBConfig;
import basics.datagraph.DataGraphManager;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.YagoClasses;

/**
 * This class is part of the YAGO converters (http://mpii.de/yago). 
 * It is licensed under the Creative Commons Attribution License 
 * (http://creativecommons.org/licenses/by/3.0)  
 * the YAGO team (http://mpii.de/yago).
 * 
 * This class inserts all text tables to an SQL Database.
 * 
 * @author Fabian M. Suchanek
 */
public class SQLConverter {

  /** Maximal entity length*/
  public static final int MAXENTITYLENGTH = 100;

  /** Holds the database */
  protected static Database database = null;

  //  /** Holds the filename of YagoStar2 (to exclude it if yagostar==false) */
  //  protected static String yagoStarFilename = "YagoStar2.txt";

  /** Special add-parameter for makeSQL */
  protected static final String INDEX = "INDEX";

  /** Tells whether the deductive closure shall be converted as well*/
  private static boolean deductiveClosure = false;

  /** Tells whether facts about facts shall be included*/
  private static boolean includeFactFacts = false;

  /** folder that contains yago */
  private static File yagoFolder = null;

  /** Adds facts from one file */
  public static void loadFacts(File file, Relation r) throws Exception {
    int[] type = { Types.INTEGER, Types.VARCHAR, Types.VARCHAR, Types.VARCHAR, Types.FLOAT };
    Inserter bulki = database.new Inserter("facts", type);
    String relation = r.name();
    for (String l : new FileLines(new UTF8Reader(file, "Parsing " + file.getName()))) {
      String[] split = l.split("\t");
      long id = Long.parseLong(split[0]);
      String arg1 = split[1];
      String arg2 = split[2];
      if (arg1.length() > MAXENTITYLENGTH || arg2.length() > MAXENTITYLENGTH) continue;
      double confidence = Double.parseDouble(split[3]);
      try {
        bulki.insert(id, relation, arg1, arg2, confidence);
      } catch (Exception e) {
        System.err.println(e);
        return;
      }
    }
    bulki.close();
  }

  /** Adds entities from one file */
  public static void loadEntities(File file) throws Exception {
    int[] type = { Types.VARCHAR, Types.INTEGER };
    Database.Inserter bulki = database.new Inserter("entities", type);
    for (String l : new FileLines(new UTF8Reader(file, "Parsing " + file.getName()))) {
      String[] split = l.split("\t");
      String entity = split[0];
      if (entity.length() > MAXENTITYLENGTH) continue;
      boolean isConcept = split[1].charAt(0) == 't';
      try {
        bulki.insert(entity, isConcept ? 1 : 0);
      } catch (Exception e) {
        System.err.println(e);
        return;
      }
    }
    bulki.close();
  }

  /** request information instead of using yago.ini file*/
  public static void requestInformation() throws Exception {
    while (yagoFolder == null) {
      D.println("Please enter a yago Folder directory containing the yago facts and entities (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      yagoFolder = new File(path);
      if (!yagoFolder.exists()) {
        D.println("yago Folder does not exist, please try again (quit to end)");
        yagoFolder = null;
      } else {
        if (!new File(yagoFolder, "facts").exists()) {
          D.println(path + " does not have a sub-folder 'facts', please try again (quit to end)");
          yagoFolder = null;
        }
        if (!new File(yagoFolder, "entities").exists()) {
          D.println(path + " does not have a sub-folder 'entities', please try again (quit to end)");
          yagoFolder = null;
        }
      }
    }
    boolean ok = false;
    while (!ok) {
      D.println("");
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
    }
    while (database == null) {
      D.println("");
      D.println("Which Database are you using (quit to end)");
      D.println("1. Oracle");
      D.println("2. MySql");
      D.println("3. Postgres");
      D.println("4. quit");
      int choice = 0;
      while (choice <= 0 || choice >= 5) {
        D.pl("input:> ");
        try {
          String input = D.r().trim();
          if (input.equals("quit")) {
            D.println("program ended");
            return;
          }
          choice = Integer.parseInt(input);
          if (choice == 4) {
            D.println("program ended");
            return;
          }
        } catch (Exception e) {
          choice = 0;
        }
      }
      String system = null;
      if (choice == 1) {
        system = "ORACLE";
        if (!OracleDatabase.checkJar()) {
          D.pl("**********************************************************************\n");
          D.pl("You are missing the oracle jdbc, see readme_converters.txt\n");
          D.pl("**********************************************************************\n");
          return;
        }
      } else if (choice == 2) {
        system = "MYSQL";
        if (!MySQLDatabase.checkJar()) {
          D.pl("**********************************************************************\n");
          D.pl("You are missing the mysql jdbc, see readme_converters.txt\n");
          D.pl("**********************************************************************\n");
          return;
        }
      } else if (choice == 3) {
        system = "POSTGRES";
        if (!PostgresDatabase.checkJar()) {
          D.pl("*************************************************************************\n");
          D.pl("You are missing the postgres jdbc, see readme_converters.txtv");
          D.pl("*************************************************************************\n");
          return;
        }
      }
      D.pl("Please enter user:> ");
      String user = D.r().trim();
      D.pl("Please enter password:> ");
      String password = D.r().trim();
      D.pl("Please enter host:> ");
      String host = D.r().trim();
      D.pl("Please enter port:> ");
      String port = D.r().trim();
      String schema = null;
      String inst = null;
      if (choice == 1) {
        D.pl("Please enter oracle SID:> ");
        inst = D.r().trim();
      }
      if (choice == 2) {
        D.pl("Please enter mysql database:> ");
        schema = D.r().trim();
      }
      if (choice == 3) {
        D.pl("Please enter postgres schema:> ");
        schema = D.r().trim();
      }
      D.println("Trying to Connect to the database");
      try {
        database = DBConfig.getDatabase(system, user, password, host, port, schema, inst);
        D.println("Connected to", database);
      } catch (Exception e) {
        D.println("Failed to Connect to database");
        database = null;
      }
    }
    String ans = null;
    while (ans == null) {
      D.println("");
      D.println("Would you like to add deductive closer of the form");
      D.println("(x,type,c1), (c1, subclassof, c2) => (x, type, c2)");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        deductiveClosure = true;
      } else if (ans.equalsIgnoreCase("n")) {
        deductiveClosure = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    ans = null;
    while (ans == null) {
      D.println("Do you want to include facts about facts");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        includeFactFacts = true;
      } else if (ans.equalsIgnoreCase("n")) {
        includeFactFacts = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    convert();
  }

  /** reads the yago.ini file*/
  public static void fromIniFile(File initFile) throws Exception {
    Parameters.init(initFile);
    DataGraphManager.init();
    Parameters.ensureParameters("yagoFolder - the YAGO data folder");
    yagoFolder = Parameters.getFile("yagoFolder");
    deductiveClosure = Parameters.isDefined("deductiveClosure") && Parameters.getBoolean("deductiveClosure");
    includeFactFacts = Parameters.isDefined("includeFactFacts") && Parameters.getBoolean("includeFactFacts");
    Announce.doing("Connecting to database");
    database = DBConfig.getDatabase(initFile.getName());
    Announce.message("Connected to", database);
    Announce.done();
    Announce.message("Deductive closure:", deductiveClosure);
    Announce.message(" fact facts:", includeFactFacts);
    convert();
  }

  public static void main(String[] args) throws Exception {
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFile(f);
    } else {
      requestInformation();
    }
  }

  /** Reads Yago to a database */
  private static void convert() throws Exception {
    D.println("Strarting to write to DB");
    // Parameter initialization
    if (!yagoFolder.exists()) Announce.error(yagoFolder, "does not exist");
    if (!new File(yagoFolder, "facts").exists()) Announce.error(yagoFolder, "does not have a sub-folder 'facts'");
    if (!new File(yagoFolder, "entities").exists()) Announce.error(yagoFolder, "does not have a sub-folder 'entities'");
    // Create facts table
    Announce.doing("Creating facts table");
    database.createTable("facts", "id", Types.INTEGER, "relation", Types.VARCHAR, "arg1", Types.VARCHAR, "arg2", Types.VARCHAR, "confidence", Types.FLOAT);
    Announce.done();

    Announce.doing("Inserting facts");
    for (File f : new File(yagoFolder, "facts").listFiles()) {
      if (!f.isDirectory()) continue;
      Relation r = Relations.get(f.getName());
      if (r == null) continue;
      if (r == Relations.CONTEXT) continue;
      if (!includeFactFacts && r.domain == YagoClasses.FACT) continue;
      Announce.doing("Inserting " + f.getName());
      for (File file : f.listFiles()) {
        if (!deductiveClosure && file.getName().equalsIgnoreCase("IsAExtractor.txt")) continue;
        if (!file.getName().endsWith(".txt")) continue;
        loadFacts(file, r);
      }
      Announce.done();
    }
    Announce.done();

    // Create entity table
    Announce.doing("Creating entity table");
    database.createTable("entities", "name", Types.VARCHAR, "isConcept", Types.INTEGER);
    Announce.done();

    // Insert entities
    Announce.doing("Inserting entities");
    for (File f : new File(yagoFolder, "entities").listFiles()) {
      loadEntities(f);
    }
    Announce.done();

    Announce.doing("Creating indices on facts");
    List<String[]> indexes = Arrays.asList(new String[] { "id" },  
        new String[] { "relation", "arg1", "arg2" }, new String[] { "relation", "arg2", "arg1" }, 
        new String[] { "arg1","relation", "arg2"}, new String[] { "arg2","relation", "arg1" },
        new String[] { "arg1","arg2","relation"}, new String[] { "arg2", "arg1","relation"});
    Announce.message("The following indexes will be created:");
    for (String[] attr : indexes)
      Announce.message(database.createIndexCommand("facts", false, attr));
    Announce.message("On some systems, this fails. In these cases, please interrupt and create the indexes manually.");
    for (String[] attr : indexes) {
      Announce.doing(database.createIndexCommand("facts", false, attr));
      database.createIndex("facts", false, attr);
      Announce.done();
    }
    Announce.done();

    Announce.doing("Creating index on entities");
    database.createIndex("entities", false, "name");
    Announce.done();

    database.close();
  }
}