package converters;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.UTF8Reader;
import javatools.filehandlers.UTF8Writer;
import javatools.parsers.Char;
import basics.datagraph.DataGraphManager;
import basics.datagraph.Relation;
import basics.datagraph.Relations;
import basics.datagraph.YagoClasses;

/**
 * This class is part of the YAGO converters (http://mpii.de/yago). 
 * It is licensed under the Creative Commons Attribution License 
 * (http://creativecommons.org/licenses/by/3.0) 
 * by the YAGO team (http://mpii.de/yago).
 * 
 * This class converts the Yago Ontology to XML
 * 
 * @author Gjergji Kasneci and Fabian M. Suchanek
 */
public class XMLConverter {

  /** Tells whether the deductive closure shall be converted as well*/
  public static boolean deductiveClosure;

  /** Tells whether facts about facts shall be included*/
  public static boolean includeFactFacts;

  /** folder that contains yago */
  private static File yagoFolder = null;

  /** target folder*/
  private static File xmlFolder = null;

  /**This method writes the corresponding dtd for the facts xml document
   * as well as the corresponding facts xml file. */
  public static void writeFactsToXML() throws IOException {
    D.println("Starting to convert YAGO to xml");
    Announce.doing("Converting facts");
    File destinationPath = new File(xmlFolder, "xmlFacts");
    destinationPath.mkdir();
    Writer dtdWriter = new BufferedWriter(new FileWriter(new File(destinationPath, "factset.dtd")));
    dtdWriter.write("<!ELEMENT factset   (relation*)>\n");
    dtdWriter.write("<!ELEMENT relation    (fact*)>\n");
    dtdWriter.write("<!ELEMENT fact      (id, arg1, arg2, confidence)>\n");
    dtdWriter.write("<!ELEMENT id          (#PCDATA)>\n");
    dtdWriter.write("<!ELEMENT arg1     (#PCDATA)>\n");
    dtdWriter.write("<!ELEMENT arg2     (#PCDATA)>\n");
    dtdWriter.write("<!ELEMENT confidence  (#PCDATA)>\n");
    dtdWriter.close();
    Writer xmlFactsWriter = new BufferedWriter(new UTF8Writer(new File(destinationPath, "xmlFacts.xml")));
    xmlFactsWriter.write("<?xml version=\"1.0\" encoding=\"UTF8\"?>\n");
    xmlFactsWriter.write("<!DOCTYPE factset SYSTEM \"factset.dtd\">\n");
    xmlFactsWriter.write("<?xml-stylesheet type=\"text\"?>\n");
    xmlFactsWriter.write("<factset>\n");
    for (File relation : new File(yagoFolder, "facts").listFiles()) {
      Relation rel = Relations.get(relation.getName());
      if (rel == null || rel.domain == YagoClasses.FACT && !includeFactFacts) continue;
      xmlFactsWriter.write(" <relation name=\"" + rel + "\">\n");
      for (File f : relation.listFiles()) {
        if (!deductiveClosure && f.getName().equalsIgnoreCase("IsAExtractor.txt")) continue;
        fileToXML(xmlFactsWriter, f, "fact", "id", "arg1", "arg2", "confidence");
      }
      xmlFactsWriter.write(" </relation>\n");
    }
    xmlFactsWriter.write("</factset>\n");
    xmlFactsWriter.close();
    Announce.done();
  }

  /**This method writes the corresponding dtd for the entities xml document
   * as well as the corresponding entities xml file. */
  public static void writeEntitiesToXML() throws IOException {
    Announce.doing("Converting entities");
    File destinationPath = new File(xmlFolder, "xmlEntities");
    destinationPath.mkdir();
    Writer dtdWriter = new BufferedWriter(new FileWriter(new File(destinationPath, "entityset.dtd")));
    dtdWriter.write("<!ELEMENT entityset   (entity*)>\n");
    dtdWriter.write("<!ELEMENT entity      (name, isConcept, description)>\n");
    dtdWriter.write("<!ELEMENT name        (#PCDATA)>\n");
    dtdWriter.write("<!ELEMENT isConcept   (#PCDATA)>\n");
    dtdWriter.write("<!ELEMENT description (#PCDATA)>\n");
    dtdWriter.close();
    Writer xmlEntityWriter = new BufferedWriter(new UTF8Writer(new File(destinationPath, "entityset.xml")));
    xmlEntityWriter.write("<?xml version=\"1.0\" encoding=\"UTF8\"?>\n");
    xmlEntityWriter.write("<!DOCTYPE entityset SYSTEM \"entityset.dtd\">\n");
    xmlEntityWriter.write("<?xml-stylesheet type=\"text\"?>\n");
    xmlEntityWriter.write("<entityset>\n");
    for (File f : new File(yagoFolder, "entities").listFiles()) {
      fileToXML(xmlEntityWriter, f, "entity", "name", "isConcept", "description");
    }
    xmlEntityWriter.write("</entityset>\n");
    xmlEntityWriter.close();
    Announce.done();
  }

  /**This method converts the facts of one YAGO file into an XML file.
   * @param out - the writer for the XML file
   * @param yagoFile - the file to read the Yago data from
   * @param frame - the tag to be put around one item
   * @param attributes - the names of the columns of the yagoFile*/
  public static void fileToXML(Writer out, File yagoFile, String frame, String... attributes) throws FileNotFoundException, IOException {
    for (String line : new FileLines(new UTF8Reader(yagoFile, "Parsing " + yagoFile.getName()))) {
      String[] split = line.split("\t");
      out.write("  <" + frame + ">\n");
      for (int i = 0; i < attributes.length; i++) {
        out.write("   <" + attributes[i] + ">");
        out.write(Char.encodeAmpersand(split[i]));
        out.write("</" + attributes[i] + ">\n");
      }
      out.write("  </" + frame + ">\n");
    }
  }

  /** reads the yago.ini file*/
  public static void fromIniFile(File initFile) throws Exception {
    Parameters.init(initFile);
    DataGraphManager.init();
    Parameters.ensureParameters("yagoFolder - the YAGO data folder", "xmlFolder - the xml target folder");
    yagoFolder = Parameters.getFile("yagoFolder");
    xmlFolder = Parameters.getFile("xmlFolder");
    if (!xmlFolder.exists()) {
      xmlFolder.mkdir();
    }
    deductiveClosure = Parameters.isDefined("deductiveClosure") && Parameters.getBoolean("deductiveClosure");
    includeFactFacts = Parameters.isDefined("includeFactFacts") && Parameters.getBoolean("includeFactFacts");
    Announce.message("Deductive closure:", deductiveClosure);
    Announce.message("Include fact facts:", includeFactFacts);
    writeFactsToXML();
  }

  /** request information instead of using yago.ini file*/
  public static void requestInformation() throws Exception {
    while (yagoFolder == null) {
      D.println("Please enter a yago Folder directory containing the yago facts and entities (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      yagoFolder = new File(path);
      if (!yagoFolder.exists()) {
        D.println("yago Folder does not exist, please try again (quit to end)");
        yagoFolder = null;
      } else {
        if (!new File(yagoFolder, "facts").exists()) {
          D.println(path + " does not have a sub-folder 'facts', please try again (quit to end)");
          yagoFolder = null;
        }
        if (!new File(yagoFolder, "entities").exists()) {
          D.println(path + " does not have a sub-folder 'entities', please try again (quit to end)");
          yagoFolder = null;
        }
      }
    }
    while (xmlFolder == null) {
      D.println("Please enter a target directory (quit to end):");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      xmlFolder = new File(path);
      if (!xmlFolder.exists()) {
        if (!xmlFolder.mkdir()) {
          xmlFolder = null;
        }
      }
    }
    boolean ok = false;
    while (!ok) {
      D.println("");
      D.println("Please enter the path for the relation and yagoclass information (quit to end)");
      D.pl("input:> ");
      String path = D.r().trim();
      if (path.equals("quit")) {
        D.println("program ended");
        return;
      }
      ok = DataGraphManager.init(path);
    }
    String ans = null;
    while (ans == null) {
      D.println("");
      D.println("Would you like to add deductive closer of the form");
      D.println("(x,type,c1), (c1, subclassof, c2) => (x, type, c2)");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        deductiveClosure = true;
      } else if (ans.equalsIgnoreCase("n")) {
        deductiveClosure = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    ans = null;
    while (ans == null) {
      D.println("Do you want to include facts about facts");
      D.println("[y]es or [n]o (quit to end)");
      D.pl("input:> ");
      ans = D.r().trim();
      if (ans.equalsIgnoreCase("y")) {
        includeFactFacts = true;
      } else if (ans.equalsIgnoreCase("n")) {
        includeFactFacts = false;
      } else if (ans.equalsIgnoreCase("quit")) {
        D.println("program ended");
        return;
      } else {
        ans = null;
      }
    }
    writeFactsToXML();
  }

  /** Converts Yago to XML*/
  public static void main(String[] args) throws Exception {
    String initFile = args == null || args.length == 0 ? "yago.ini" : args[0];
    File f = new File(initFile);
    if (f.exists()) {
      fromIniFile(f);
    } else {
      requestInformation();
    }
  }
}
